#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__all__ = ['BaSiC']

from pathlib import Path
import cv2
import numpy as np
from scipy.fftpack import dctn, idctn
import tqdm
from numba import njit

# From:
# linum-uqam/PyBaSiC: v1.0.0
# Joël Lefebvre
# 10.5281/zenodo.7305570
# https://github.com/linum-uqam/PyBaSiC

class BaSiC(object):
    def __init__(self, input, estimate_darkfield=False, extension=".tif", verbose=False,
                 working_size=128, epsilon=0.1, l_s=None, l_d=None, reweighting_tolerance=1e-3,
                 max_reweightingIterations=10):
        """Input can either be:
         - Path to a directory containing the images to process
         - List of images path to process
         - List of images as ndarray
         - A stack of ndarrays of shape N_Images x Height x Width
        """
        self.input_type = None
        self.extension = extension
        if isinstance(input, str) or isinstance(input, Path): # Directory
            self.directory = input
            self._sniff_input() # Get a list of files
            self.input_type = "directory"
        elif isinstance(input, np.ndarray):
            self.img_stack = input
            self.input_type = "images_stack"
        elif isinstance(input, list) and (isinstance(input[0], str) or isinstance(input[0], Path)):
            self.files = input
            self.input_type = "files_list"
        elif isinstance(input, list) and isinstance(input[0], np.ndarray):
            self.img_stack = np.array(input)
            self.input_type = "images_list"
        else:
            raise "input should either be a directory, a list of ndarrays, or a ndarray stack."

        # Optimizer parameters
        self.working_size = working_size  # px : image resampling size to accelerate learning.
        self.epsilon = epsilon  # Iterative reweighted L1-norm stability parameter
        self.l_s = l_s  # flat-field regularization parameter (set automatically if None)
        self.l_d = l_d  # dark-field regularization parameter (set automatically if None)
        self.reweighting_tolerance = reweighting_tolerance
        self.max_reweightingIterations = max_reweightingIterations
        self.reweighting_iteration = 0
        self.estimate_darkfield = estimate_darkfield
        self.verbose = verbose

    def _sniff_input(self):
        # Get a list of tiles to process
        directory = Path(self.directory).resolve()
        file_list = list(directory.glob(f"*{self.extension}"))
        file_list.sort()
        self.files = file_list
        assert len(self.files) > 0, "No files were found in the input directory. Make sure you provided the right path and file extension. Aborting."

    def _load_images(self, img_stack = None):
        # Load the stack
        if img_stack is None:
            img_stack = []
            if self.verbose:
                gen = tqdm.tqdm(self.files, "Loading the images")
            else:
                gen = self.files
            for this_file in gen:
                img = cv2.imread(str(this_file), cv2.IMREAD_ANYDEPTH)
                img_stack.append(img)
            self.img_stack = np.array(img_stack)
        else:
            self.img_stack = img_stack
        self.n_images = self.img_stack.shape[0]

        # Resample the images to accelerate learning
        self.image_shape = self.img_stack.shape[1::]
        new_shape = tuple([self.working_size]*2)
        img_stack_p = np.zeros([self.n_images, *new_shape], dtype=self.img_stack.dtype)
        if self.working_size > self.image_shape[0]:
            interpolation = cv2.INTER_LINEAR
        else:
            interpolation = cv2.INTER_AREA
        for i in range(self.n_images):
            img = self.img_stack[i, ...].squeeze()
            img_stack_p[i, ...] = cv2.resize(img.T, new_shape, interpolation=interpolation).T
        self.img_stack_resized = img_stack_p.astype(np.float32)

    def normalize(self, img, clip=True, epsilon=1e-6):
        img_p = (img.astype(np.float32) - self.darkfield_fullsize) / (self.flatfield_fullsize + epsilon)
        if clip and not(img.dtype in [np.float32, np.float64]):

            img_p[img_p < np.iinfo(img.dtype).min] = np.iinfo(img.dtype).min
            img_p[img_p > np.iinfo(img.dtype).max] = np.iinfo(img.dtype).max

        return img_p.astype(img.dtype)

    def write_images(self, directory, epsilon=1e-6):
        # Create the output directory
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Loop over the images
        for i in tqdm.tqdm(range(self.n_images), desc="Shading Correction"):
            this_file = self.files[i]
            this_img = self.img_stack[i, ...]

            # Normalize the image
            img_p = self.normalize(this_img, epsilon=epsilon)

            # Get the output filename
            filename = directory / this_file.name

            # Save the file
            cv2.imwrite(str(filename), img_p)

    def prepare(self, img_stack=None):
        # Load the data
        if img_stack is not None:
            self._load_images(img_stack)
        elif self.input_type in ["directory", "files_list"]:
            self._load_images()
        elif self.input_type in ["images_stack", "images_list"]:
            self._load_images(self.img_stack)

        # Initialize the regularization parameters
        mean_value = self.img_stack_resized.mean(axis=0)
        mean_value /= mean_value.mean() # Normalized pixel-wise mean of all images
        mean_value_dct = dctn(mean_value, norm='ortho')

        if self.l_s is None:
            self.l_s = np.abs(mean_value_dct).sum() / 800.0
        if self.l_d is None:
            self.l_d = np.abs(mean_value_dct).sum() / 2000.0

        # Construct the measurement matrix
        self.img_sort = np.sort(self.img_stack_resized, axis=0)

        # Initialize the darkfield, flatfield, offset, and weights (for the L1 reweighted loss)
        self.offset = np.zeros([self.working_size]*2)
        self.flatfield = np.ones([self.working_size]*2)
        self.flatfield_fullsize = np.ones(self.image_shape)
        self.darkfield = np.random.randn(self.working_size, self.working_size)
        self.darkfield_fullsize = np.zeros(self.image_shape)
        self.W = np.ones_like(self.img_sort)

        # Initialize other parameters
        self.iteration = 0
        self.flag_reweigthing = True

    def update_weights(self):
        """Weighting matrix for the Reweighted L1-norm"""
        # Weight Update formula in the paper
        # self.W = 1.0 / (np.abs(self.Ir / self.Ib) + self.epsilon)

        # Actual Weight update formula in the matlab implementation
        self.W = 1.0 / (np.abs(self.Ir / (self.Ib.mean() + 1e-6)) + self.epsilon)
        self.W = self.W * self.W.size / self.W.sum()
        self.reweighting_iteration += 1

    def update(self):
        last_flatfield = self.flatfield.copy()
        last_darkfield = self.darkfield.copy()

        # Perform LADM optimization
        Ib, Ir, D = inexact_alm_l1(self.img_sort, self.l_s, self.l_d, weight=self.W, estimateDarkField=self.estimate_darkfield, verbose=self.verbose)

        # Reshape the images.
        self.Ib = np.reshape(Ib, (self.n_images, self.working_size, self.working_size)) # Flat-field
        self.Ir = np.reshape(Ir, (self.n_images, self.working_size, self.working_size)) # Residual
        D = np.reshape(D, (self.working_size, self.working_size)) # Dark-field

        # Update the weight matrix
        self.update_weights()

        # Update the flat-field and dark-field
        self.flatfield = self.Ib.mean(axis=0) - D
        self.flatfield = self.flatfield / self.flatfield.mean()
        self.darkfield = D

        # Compute the difference between the new fields and the last ones.
        mad_flatfield = np.abs(self.flatfield - last_flatfield).sum() / np.abs(last_flatfield).sum()
        mad_darkfield = np.abs(self.darkfield - last_darkfield).sum()
        if mad_darkfield < 1e-7:
            mad_darkfield = 0
        else:
            mad_darkfield = mad_darkfield / max(np.abs(last_darkfield).sum(), 1e-6)

        # Check if another L1 reweighting is necessary
        if (max(mad_flatfield, mad_darkfield)<=self.reweighting_tolerance) or (self.reweighting_iteration > self.max_reweightingIterations):
            self.flag_reweigthing = False

    def run(self):
        if self.verbose:
            pbar = tqdm.tqdm(desc="Reweighting Iteration", total=self.max_reweightingIterations)
        while self.flag_reweigthing:
            self.update()
            if self.verbose:
                pbar.update()
            # self.display_fields()
        if self.verbose:
            pbar.close()

        # Reshape the flat and dark fields to the original shape
        self.flatfield_fullsize = cv2.resize(self.flatfield.T, self.image_shape, cv2.INTER_LINEAR).T
        self.flatfield_fullsize = self.flatfield_fullsize / self.flatfield_fullsize.mean()
        self.darkfield_fullsize = cv2.resize(self.darkfield.T, self.image_shape, cv2.INTER_LINEAR).T

    def set_flatfield(self, flatfield):
        self.flatfield_fullsize = cv2.resize(flatfield.T, self.image_shape, cv2.INTER_LINEAR).T

    def set_darkfield(self, darkfield):
        self.darkfield_fullsize = cv2.resize(darkfield.T, self.image_shape, cv2.INTER_LINEAR).T

    def get_flatfield(self):
        return self.flatfield_fullsize.copy()

    def get_darkfield(self):
        return self.darkfield_fullsize.copy()

def shrink(theta, epsilon=1e-3):
    """Scalar Shrink Operator"""
    theta_p = np.sign(theta) * np.maximum(np.abs(theta) - epsilon, 0)
    return theta_p

def inexact_alm_l1(imgs, l_s, l_d, tol=1e-6, maxIter=500, weight=1, estimateDarkField=True, rho=1.5, verbose=False):
    """l1 minimization using the inexact augmented Lagrange multiplier method for Sparse low rank matrix recovery.
    Parameters
    ----------
    imgs : N x P x Q ndarray
        Images stack
    l_s : float
        Flat-field regularization parameter
    l_d : float
        Dark-field regularization parameter
    tol : float
        Convergence tolerance
    maxIter : int
        Maximum iterations number
    weight : N x P x Q ndarray
        Optional weight matrix used for the reweighted L1 norm
    estimateDarkField : bool
        Set to True to estimate the darkfield in addition to the flat field
    darkFieldLimit : float
        Maximum value for the darkfield, use to constrain the minimization
    rho : float
        Lagrange multiplier learning rate

    Returns
    -------
    S : ndarray
        Estimated unnormalised flat-field
    Ib : ndarray

    Ir : ndarray



    Notes
    -----
    % modified from the BaSiC matlab implementation
    % modified from Robust PCA
    % reference:
    % Peng et al. "A BaSiC tool for background and shading correction
    % of optical microscopy images" Nature Communications, 14836(2017)
    % Candès, E., Li, X., Ma, Y. & Wright, J. "Robust Principal Component
    % Analysis?" J. ACM (58) 2011

    % D - m x m x n matrix of observations/data (required input)
    %

    % while ~converged
    %   minimize (inexactly, update A and E only once)
    %   L(W, E,Y,u) = |E|_1+lambda * |W|_1 + <Y2,D-repmat(QWQ^T)-E> + +mu/2 * |D-repmat(QWQ^T)-E|_F^2;
    %   Y1 = Y1 + \mu * (D - repmat(QWQ^T) - E);
    %   \mu = \rho * \mu;
    % end
    %
    % Tingying Peng (tingying.peng@tum.de)

    %
    % Copyright: CAMP, Technical University of Munich

    """
    ###############################
    # Initialize the optimization #
    ###############################
    N, P, Q = imgs.shape[:]

    # Reshape the image stack as a measurement matrix
    D = np.reshape(imgs, (N, P*Q))
    d_norm = np.linalg.norm(D, "fro")
    W = np.reshape(weight, D.shape)
    B1_uplimit = D.min()
    B1 = 0

    # Initialize Ib, Ir, B, S and D
    S = np.zeros_like(D)  # Flat-field
    Sf = dctn(np.reshape(S, (N,P,Q)).mean(axis=0), norm='ortho') # Flat-field, in Fourier Domain
    Ir = np.zeros_like(D)  # Residual
    B = np.ones((N,1)) # Image baseline
    D_field = np.zeros((1, P*Q)) # Darkfield

    # Initialize Lagrange multiplier
    Y = 0
    D_svd = np.linalg.svd(D, compute_uv=False)
    mu = 12.5 / D_svd[0] # TODO: This can be tuned
    mu_bar = mu * 1e7
    ent2 = 10

    converged = False
    iteration = 0
    if verbose:
        pbar = tqdm.tqdm(desc="Iteration", total=maxIter)
    while not(converged) and (iteration < maxIter): # TODO: Add tqdm
        # 1. Compute DCT of the flat-field Sf(k), and Ib(k)
        S = np.reshape(idctn(Sf, norm='ortho'), (1,P*Q)) # Flat-field in spatial domain
        Ib = S * B + D_field # with broadcasting, as S (NxPQ), B (Nx1), D(1xPQ)

        # 2. Update Sf(k) -> Sf(k+1)
        dS = (D - Ib - Ir + Y / mu) # Shape: N x PQ
        dS = np.reshape(dS, (N, P, Q))
        dS = dS.mean(axis=0) # Averate over all images
        Sf = Sf + dctn(dS, norm='ortho')
        Sf = shrink(Sf, l_s / mu) # Scalar shrinkage operator

        # 3. Get S(k+1,x) and mid-iteration Ib(k+1/2)
        S = np.reshape(idctn(Sf, norm='ortho'), (1, P*Q)) # Flat-field at iteration S(k+1)
        Ib = S * B + D_field # Ib, mid iteration

        # 4. Update the residual Ir(k) -> Ir(k+1)
        dIr = (D - Ib - Ir + Y / mu)
        Ir = Ir + dIr
        Ir = shrink(Ir, W/mu)

        # 5. Update the image baseline B(k) -> B(k+1), and Ib(k+1)
        R = D - Ir
        B = R.mean(axis=1, keepdims=True) / R.mean()
        B[B < 0] = 0 # Enforce non-negativity constraint
        # Ib = S * B + D_field

        # Dark-field optimization
        if estimateDarkField: # TODO: Add the darkfield optimization here
            # 1. Dz estimation with least-mean square
            mask_validB = B < 1
            mask_highS = S > (S.mean() - 1e-6)
            mask_lowS = S < (S.mean() + 1e-6)
            R_high = np.mean(R * (mask_highS * mask_validB), axis=1, keepdims=True)
            R_low = np.mean(R * (mask_lowS * mask_validB), axis=1, keepdims=True)
            B1 = (R_high - R_low)/R.mean()
            k = mask_validB.sum()
            temp1 = np.sum(B[mask_validB]**2)
            temp2 = B[mask_validB].sum()
            temp3 = B1.sum()
            temp4 = np.sum(B[mask_validB] * B1)
            temp5 = temp2 * temp3 - k * temp4
            if temp5 == 0:
                B1 = 0
            else:
                B1 = (temp1*temp3 - temp2*temp4) / temp5

            # Clipping to limit the range of Dz to 0 and min(D)
            B1 = np.maximum(B1, 0) # Non negativity constraing
            B1 = np.minimum(B1, B1_uplimit / S.mean())

            # 2. Dr optimization
            Z = B1 * (S.mean() - S)

            A1_offset = np.ma.masked_array(R, np.tile(~mask_validB, (1, P*Q))).mean(axis=0, keepdims=True) - B[mask_validB].mean() * S
            A1_offset = A1_offset - A1_offset.mean()
            A_offset = A1_offset - A1_offset.mean() - Z

            # Smooth A_offset (Dr)
            Dr_f = dctn(np.reshape(A_offset, (P,Q)), norm='ortho') # Fourier-domain flat-field residual
            Dr_f = shrink(Dr_f, l_d / (ent2 * mu)) # Shrink operator in Fourier-domain for smooth residual
            Dr = idctn(Dr_f, norm='ortho').reshape((1, P*Q))
            Dr = shrink(Dr, l_d / (mu*ent2)) # Shrink operator in spatial domain for sparse redisual
            D_field = Dr + Z

        # 6. Update the Lagrange multiplier Y(k)->Y(k+1), mu(k)->mu(k+1), and k->k+1
        dY = D - Ib - Ir
        Y = Y + mu * dY
        mu = min(mu * rho, mu_bar) # Clip mu
        iteration += 1

        # Evaluate convergence
        stopCriterion = np.linalg.norm(dY, "fro") / d_norm
        if verbose:
            pbar.update()
        # pbar.set_description(f"Iteration (Criterion={stopCriterion:.3e})", refresh=False)
        if stopCriterion < tol:
            converged = True

        # TODO: Add logging

    if iteration == maxIter:
        print("Maximum iterations reached")
    if verbose:
        pbar.close()
    # Update the darkfield, with
    D_field = D_field + B1*S

    return Ib, Ir, D_field










@njit
def mean_axis_0(arr):
    return np.array([arr[..., i].mean() for i in range(arr.shape[1])])

def get_photobleach(imgflt_stack, flatfield, darkfield=None, size=(128,128)):
    # Size is (x,y)
    img_stack = np.zeros((size[1], size[0], imgflt_stack.shape[0]))

    for i, frame in enumerate(imgflt_stack):
        img_stack[:, :, i] = cv2.resize(frame, size, interpolation=cv2.INTER_LINEAR).astype(np.float64)

    imgflt_stack = img_stack

    flatfield_small = cv2.resize(flatfield, size, interpolation=cv2.INTER_LINEAR).astype(np.float64)
    if darkfield is not None:
        darkfield_small = cv2.resize(darkfield, size, interpolation=cv2.INTER_LINEAR).astype(np.float64)

    imgflt_stack = np.reshape(imgflt_stack, (size[0] * size[1], -1)).astype(np.float64)

    imgflt_stack_svd = np.linalg.svd(imgflt_stack, full_matrices=False, compute_uv=False)

    return _get_photobleach(imgflt_stack, imgflt_stack_svd, flatfield_small, darkfield_small)

# From https://github.com/PolusAI/polus-plugins/tree/dev/regression/polus-basic-flatfield-correction-plugin
# ------------------------------------------------------------------------------
# MIT License
#
# Copyright (c) 2019 LabShare
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ------------------------------------------------------------------------------

@njit
def _get_photobleach(imgflt_stack, imgflt_stack_svd, flatfield, darkfield=None):
    """Calculate the global effect of photobleaching for each image
    Using the original data, flatfield, and darkfield images, estimate the total
    contribution of photobleaching to an image in a series of images.
    Inputs:
        imgflt_stack - Numpy stack of images
        flatfield - numpy floating precision matrix containing flatfield values
        darkfield - numpy floating precision matrix containing darkfield values
    Outputs:
        A_coeff - A 1xn matrix of photobleaching offsets, where n is the number
            of input images
    """


    # Initialize matrices
    # Initialize matrices
    # imgflt_stack = np.reshape(
    #     imgflt_stack, (size * size, -1)
    # ).astype(np.float64)
    if darkfield is None:
        darkfield = np.zeros(flatfield.shape, dtype=np.float64)

    # Initialize weights and tolerances
    weights = np.ones(imgflt_stack.shape, dtype=np.float64)
    epsilon = np.float64(0.1)
    tol = np.float64(10 ** -6)

    # Run optimization exactly 5 times
    for r in range(5):
        # print(r)
        # Calculate weights, offsets and coefficients
        W_idct_hat = np.reshape(flatfield, (-1, 1))
        A_offset = np.reshape(darkfield, (-1, 1))
        A_coeff = np.reshape(mean_axis_0(imgflt_stack), (1, -1))

        # Initialization values and learning rates
        # temp = np.linalg.svd(imgflt_stack, full_matrices=False, compute_uv=False)
        temp = imgflt_stack_svd
        norm_two = np.float64(temp[0])
        mu = np.float64(12.5) / norm_two
        mu_bar = mu * 10 ** 7
        rho = np.float64(1.5)

        ent1 = 1

        # Normalization factors
        d_norm = np.linalg.norm(imgflt_stack)#, "fro")

        # Initialize augmented representation and error
        A = np.zeros(imgflt_stack.shape, dtype=np.float64)
        E1 = np.zeros(imgflt_stack.shape, dtype=np.float64)
        # Y1 = np.float64(0)
        Y1 = np.zeros(imgflt_stack.shape)

        # Run optimization
        iternum = 0
        converged = False
        while not converged:
            iternum += 1
            # print(iternum)
            # Calculate augmented representation
            # A = np.matmul(W_idct_hat, A_coeff) + A_offset
            A = (W_idct_hat * A_coeff) + A_offset

            # Calculate errors
            # E1 = E1 + np.divide(imgflt_stack - A - E1 + np.multiply(1 / mu, Y1), ent1)
            E1 = imgflt_stack - A + Y1 / mu

            # E1 = np.max(
            #     np.reshape(
            #         E1 - weights / (ent1 * mu),
            #         (imgflt_stack.shape[0], imgflt_stack.shape[1], 1),
            #     ),
            #     -1,
            #     initial=10 ** -6,
            # ) + np.min(
            #     np.reshape(
            #         E1 + weights / (ent1 * mu),
            #         (imgflt_stack.shape[0], imgflt_stack.shape[1], 1),
            #     ),
            #     -1,
            #     initial=0,
            # )

            wem = weights / (ent1 * mu)
            E1 = np.maximum(E1 - wem, 10 ** -6) + np.minimum(E1 + wem, 0)

            # Calculate coefficients
            R1 = imgflt_stack - E1
            # A_coeff = np.reshape(np.mean(R1, axis=0), (1, -1)) - np.mean(A_offset)
            A_coeff = np.reshape(mean_axis_0(R1), (1, -1)) - np.mean(A_offset)
            # A_coeff[A_coeff < 0] = 0  # pixel values are never negative
            A_coeff = np.maximum(A_coeff, 0)

            # Loss
            Z1 = imgflt_stack - A - E1

            # Error updates
            Y1 = Y1 + mu * Z1

            # Update learning rate
            # mu = np.min(mu * rho, initial=mu_bar)
            mu = np.minimum(mu * rho, mu_bar)

            # Stop if below threshold
            #stopCriterion = np.linalg.norm(Z1, "fro") / d_norm
            stopCriterion = np.linalg.norm(Z1) / d_norm

            # print(stopCriterion)
            if stopCriterion < tol:
                converged = True

        # Update weights
        # XE_norm = np.reshape(np.mean(A, axis=0), (1, -1)) / E1
        XE_norm = np.reshape( mean_axis_0(A), (1, -1)) / E1
        weights = 1 / np.abs(XE_norm + epsilon)
        weights = weights * weights.size / np.sum(weights)

    return A_coeff


#
# #original
#
# OPTIONS = {'size':128}
# def _get_photobleach(imgflt_stack, flatfield, darkfield=None):
#     """Calculate the global effect of photobleaching for each image
#     Using the original data, flatfield, and darkfield images, estimate the total
#     contribution of photobleaching to an image in a series of images.
#     Inputs:
#         imgflt_stack - Numpy stack of images
#         flatfield - numpy floating precision matrix containing flatfield values
#         darkfield - numpy floating precision matrix containing darkfield values
#     Outputs:
#         A_coeff - A 1xn matrix of photobleaching offsets, where n is the number
#             of input images
#     """
#     # Initialize matrices
#     imgflt_stack = np.reshape(
#         imgflt_stack, (OPTIONS["size"] * OPTIONS["size"], -1)
#     ).astype(np.float64)
#     if darkfield is None:
#         darkfield = np.zeros(flatfield.shape, dtype=np.float64)
#
#     # Initialize weights and tolerances
#     weights = np.ones(imgflt_stack.shape, dtype=np.float64)
#     epsilon = np.float64(0.1)
#     tol = np.float64(10 ** -6)
#
#     # Run optimization exactly 5 times
#     for r in range(5):
#         # Calculate weights, offsets and coefficients
#         W_idct_hat = np.reshape(flatfield, (-1, 1))
#         A_offset = np.reshape(darkfield, (-1, 1))
#         A_coeff = np.reshape(np.mean(imgflt_stack, 0), (1, -1))
#
#         # Initialization values and learning rates
#         temp = np.linalg.svd(imgflt_stack, full_matrices=False, compute_uv=False)
#         norm_two = np.float64(temp[0])
#         mu = np.float64(12.5) / norm_two
#         mu_bar = mu * 10 ** 7
#         rho = np.float64(1.5)
#         ent1 = 1
#
#         # Normalization factors
#         d_norm = np.linalg.norm(imgflt_stack, "fro")
#
#         # Initialize augmented representation and error
#         A = np.zeros(imgflt_stack.shape, dtype=np.float64)
#         E1 = np.zeros(imgflt_stack.shape, dtype=np.float64)
#         Y1 = np.float64(0)
#
#         # Run optimization
#         iternum = 0
#         converged = False
#         while not converged:
#             iternum += 1
#
#             # Calculate augmented representation
#             A = np.matmul(W_idct_hat, A_coeff) + A_offset
#
#             # Calculate errors
#             E1 = E1 + np.divide(imgflt_stack - A - E1 + np.multiply(1 / mu, Y1), ent1)
#             E1 = np.max(
#                 np.reshape(
#                     E1 - weights / (ent1 * mu),
#                     (imgflt_stack.shape[0], imgflt_stack.shape[1], 1),
#                 ),
#                 -1,
#                 initial=10 ** -6,
#             ) + np.min(
#                 np.reshape(
#                     E1 + weights / (ent1 * mu),
#                     (imgflt_stack.shape[0], imgflt_stack.shape[1], 1),
#                 ),
#                 -1,
#                 initial=0,
#             )
#
#             # Calculate coefficients
#             R1 = imgflt_stack - E1
#             A_coeff = np.reshape(np.mean(R1, 0), (1, -1)) - np.mean(A_offset)
#             A_coeff[A_coeff < 0] = 0  # pixel values are never negative
#
#             # Loss
#             Z1 = imgflt_stack - A - E1
#
#             # Error updates
#             Y1 = Y1 + mu * Z1
#
#             # Update learning rate
#             mu = np.min(mu * rho, initial=mu_bar)
#
#             # Stop if below threshold
#             stopCriterion = np.linalg.norm(Z1, "fro") / d_norm
#             if stopCriterion < tol:
#                 converged = True
#
#         # Update weights
#         XE_norm = np.reshape(np.mean(A, 0), (1, -1)) / E1
#         weights = 1 / np.abs(XE_norm + epsilon)
#         weights = weights * weights.size / np.sum(weights)
#
#     return A_coeff


# Implemented based on Matlab code https://github.com/marrlab/BaSiC
# def BaSiC_basefluor(IF, flatfield, darkfield=None):
#     img_stack = IF
#     nrows = 128
#     ncols = 128
#     working_size = 128
#     n_images = img_stack.shape[0]
#
#
#     image_shape = img_stack.shape[1::]
#     new_shape = tuple([working_size] * 2)
#     img_stack_p = np.zeros([n_images, *new_shape], dtype=img_stack.dtype)
#     if working_size > image_shape[0]:
#         interpolation = cv2.INTER_LINEAR
#     else:
#         interpolation = cv2.INTER_AREA
#     for i in range(n_images):
#         img = img_stack[i, ...].squeeze()
#         img_stack_p[i, ...] = cv2.resize(img.T, new_shape, interpolation=interpolation).T
#     img_stack_resized = img_stack_p.astype(np.float32)
#     D = img_stack_resized.reshape(-1,nrows*ncols)
#
#
#     flatfield = cv2.resize(flatfield.T, new_shape, interpolation=interpolation).T
#     if darkfield is not None:
#         darkfield = cv2.resize(darkfield.T, new_shape, interpolation=interpolation).T
#
#     weight = np.ones(D.shape)
#     eplson = 0.1
#     tol = 1e-6
#
#     for reweighting_iter in range(5):
#         W_idct_hat = flatfield.flatten()
#         A_offset = darkfield.flatten()
#         A1_coeff = np.mean(D, 0)
#         # main iteration loop starts
#         temp = np.linalg.svd(D, compute_uv=False)
#         norm_two = temp[0]
#         mu = 12.5 / norm_two # this one can be tuned
#         mu_bar = mu * 1e7
#         rho = 1.5 # this one can be tuned
#         d_norm = np.linalg.norm(D, 'fro')
#         ent1 = 1
#         iter = 0
#         total_svd = 0
#         converged = False
#         A1_hat = np.zeros(D.shape)
#         E1_hat = np.zeros(D.shape)
#         Y1 = 0
#
#         while not converged:
#             iter = iter + 1
#             A1_hat = W_idct_hat * A1_coeff + A_offset
#
#             # update E1 using l0 norm
#             E1_hat = E1_hat + (D - A1_hat - E1_hat + (1 / mu) * Y1) / ent1
#             E1_hat = np.maximum(E1_hat - weight / (ent1 * mu), 0) + np.minimum(E1_hat + weight / (ent1 * mu), 0)
#             # update A1_coeff, A2_coeff and A_offset
#             # if coeff_flag
#
#             # R1 = bsxfun( @ minus, D - E1_hat, A_offset)
#             R1 = D - E1_hat - A_offset
#             A1_coeff = np.mean(R1,0) - np.mean(A_offset)
#             # [A1_mincoeff, A1_mincoefftile] = min(A1_coeff);
#             # A1_back(A1_coeff < 1) = 1;
#             A1_coeff[A1_coeff < 0] = 0
#
#             # A_offset = max(min(A_offset, A_uplimit), 0);
#             Z1 = D - A1_hat - E1_hat
#             # Z2 = D - A2_hat - E2_hat;
#
#             Y1 = Y1 + mu * Z1
#             # Y2 = Y2 + mu * Z2;
#
#             mu = np.minimum(mu * rho, mu_bar)
#
#             # stop Criterion
#             stopCriterion = np.linalg.norm(Z1, 'fro') / d_norm
#             # stopCriterion = max(norm(Z1, 'fro') / d_norm, norm(Z2, 'fro') / d_norm);
#             if stopCriterion < tol:
#                 converged = True
#
#             if total_svd % 10 == 0:
#                 print('Iteration ' + str(iter) + ' |E1|_0 ' + str(np.sum(np.abs(E1_hat.flatten()) > 0)) + \
#                 ' stopCriterion ' + str(stopCriterion))
#
#         # update weight
#         # XE_norm = bsxfun( @ ldivide, E1_hat, mean(A1_hat));
#         XE_norm = np.mean(A1_hat) / E1_hat
#         weight = 1 / (np.abs(XE_norm) + eplson)
#
#         # % if isempty(segmentation)
#         #     % weight(segmentation
#         #     ~ = 0)=0;
#         # % end
#         weight = weight * weight.size / weight.sum()
#
#     return A1_coeff
#



import math
import numpy as np
import xarray as xr
from pathlib2 import Path


###################################################
## To enable interactive plotting with PySide2 in PyCharm 2022.3
import PySide2
import sys
sys.modules['PyQt5'] = sys.modules['PySide2']
from matplotlib import use
use('Qt5Agg')
###################################################

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.patches import ConnectionPatch
from mpl_toolkits.axes_grid1 import make_axes_locatable




# ## Generated dataset
# sequence = 'AAAA'
# sequences = [sequence]
# bases = 'ACTG'
# for i in range(len(sequence)):
#     for base in bases:
#         new = sequence[0:i] + base + sequence[i + 1:len(sequence)]
#         if new != sequence:
#             sequences.append(new)
# da = xr.DataArray(np.random.rand(len(sequences)), coords={'sequence': sequences})


def single_mutations(sequence):
    sequences = [] # [sequence]
    bases = 'ACTG'
    for i in range(len(sequence)):
        for base in bases:
            new = sequence[0:i] + base + sequence[i + 1:len(sequence)]
            if new != sequence:
                sequences.append(new)
    return sequences

def plot_single_mutation_data(sequence, da):
    sequences = [sequence] + single_mutations(sequence)
    da = da.sel(sequence=sequences)

    fig, ax = plt.subplots(figsize=(1+len(sequence)*0.75, 3), tight_layout=True)
    width = 0.25
    position = np.arange(len(sequence))
    x = np.hstack([-1, np.vstack([position - width, position, position + width]).T.flatten()])
    ax.bar(x, da.values, width=width)

    ticklabels = ['WT'] + [label for p in position for label in list(bases.replace(sequence[p], ''))]
    ax.set_xticks(x, ticklabels)
    ax.set_ylabel(da.name)

    for p in position:
        # for xi in x[p*3+1:(p+1)*3+1]:
        # ax.annotate(seq[p], xy=(xi,-20), xytext=(p,-20), xycoords=('data', 'axes points'), textcoords=('data','offset points'),
        #             ha='center', va='bottom', arrowprops={'width': 0.3, 'headwidth': 3, 'headlength':2})
        # text = seq[p]
        for xi in [-1, 0, 1]:
            if xi == 0:
                text = sequence[p]
            else:
                text = ' '
            ax.annotate(text, xy=(p + xi * width * 2.3 / 3, -17), xytext=(p, -20), xycoords=('data', 'axes points'),
                        textcoords=('data', 'offset points'),
                        ha='center', va='bottom', arrowprops={'width': 0.2, 'headwidth': 3, 'headlength': 2})

def double_mutations(sequence, add_reference=False):
    ## Generated dataset
    # sequence = 'GGCGCCGC'
    sequences = []
    if add_reference:
        sequences.append(sequence)
    bases = 'ACTG'

    for i in range(len(sequence)):
        for j in range(len(sequence)):
            if i == j:
                for base in bases:
                    if base != sequence[i]:
                        print(i, j, base)
                        new = sequence[0:i] + base + sequence[(i+1):len(sequence)]
                        sequences.append(new)
            elif i < j:
                for base1 in bases:
                    for base2 in bases:
                        if base1 != sequence[i] and base2 != sequence[j]:
                            print(i, j, base1, base2)
                            new = sequence[0:i] + base1 + sequence[(i+1):j] + base2 + sequence[(j + 1):len(sequence)]
                            sequences.append(new)

    return sequences


def plot_double_mutations(sequence, da, da_annotation=None, save=False, save_path=None, **kwargs):
    # sequences = [sequence] + double_mutations(sequence)
    bases = 'ACTG'
    data = np.zeros((len(sequence)*3, len(sequence)*3))
    data_annotation = np.zeros((len(sequence)*3, len(sequence)*3))
    data[:] = np.nan
    data_annotation[:] = np.nan
    bases_axes = []

    for i in range(len(sequence)):
        for j in range(len(sequence)):
            if i == j:
                k = 0
                for base in bases:
                    if base != sequence[i]:
                        new = sequence[0:i] + base + sequence[(i+1):len(sequence)]
                        # sequences.append(new)
                        data[i*3+k,i*3+k] = da.sel(sequence=new).item()
                        if da_annotation is not None:
                            data_annotation[i * 3 + k, i * 3 + k] = da_annotation.sel(sequence=new).item()
                        bases_axes.append(base)
                        # print(i, j, k, base)
                        k += 1
            elif i < j:
                k1 = 0
                for base1 in bases:
                    if base1 != sequence[i]:
                        k2 = 0
                        for base2 in bases:
                             if base2 != sequence[j]:
                                new = sequence[0:i] + base1 + sequence[(i+1):j] + base2 + sequence[(j + 1):len(sequence)]
                                # sequences.append(new)
                                data[j*3+k2, i*3+k1] = da.sel(sequence=new).item()
                                if da_annotation is not None:
                                    data_annotation[j * 3 + k2, i * 3 + k1] = da_annotation.sel(sequence=new).item()
                                # print(i, j, k1, k2, base1, base2)
                                k2 += 1
                        k1 += 1

    figsize = 1+len(sequence)*0.75
    fig, ax = plt.subplots(figsize=(figsize, figsize), tight_layout=True)

    if sequence in da.sequence:
        data[0,data.shape[1]//2] = da.sel(sequence=sequence).item()
        if da_annotation is not None:
            data_annotation[0, data.shape[1] // 2] = da_annotation.sel(sequence=sequence).item()

    # Plot the heatmap
    im = ax.imshow(data, cmap="Greens", **kwargs)


    # Create colorbar
    cax = ax.inset_axes([1.04, 0.2, 0.05, 0.6], transform=ax.transAxes)
    cbar = ax.figure.colorbar(im, ax=ax, cax=cax)
    cbar.ax.set_ylabel(da.name.capitalize().replace('_',' '), rotation=-90, va="bottom")

    # Show all ticks and label them with the respective list entries.
    ax.set_xticks(np.arange(data.shape[1]), labels=bases_axes)
    ax.set_yticks(np.arange(data.shape[0]), labels=bases_axes)
    #
    # # Let the horizontal axes labeling appear on top.
    # ax.tick_params(top=True, bottom=False,
    #                labeltop=True, labelbottom=False)
    #
    # # Rotate the tick labels and set their alignment.
    # plt.setp(ax.get_xticklabels(), rotation=-30, ha="right",
    #          rotation_mode="anchor")

    # Turn spines off and create white grid.
    ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(0, data.shape[1] + 1, 3) - .5, minor=True)
    ax.set_yticks(np.arange(0, data.shape[0] + 1, 3) - .5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)
    ax.set_title(sequence)


    position = np.arange(len(sequence))
    # x = np.hstack([np.vstack([position-1, position, position+1]).T.flatten()])

    for p in position:
        #for xi in x[p*3+1:(p+1)*3+1]:
            # ax.annotate(seq[p], xy=(xi,-20), xytext=(p,-20), xycoords=('data', 'axes points'), textcoords=('data','offset points'),
            #             ha='center', va='bottom', arrowprops={'width': 0.3, 'headwidth': 3, 'headlength':2})
        text = sequence[p]
        for xi in [-1,0,1]:
            if xi == 0:
                text = sequence[p]+str(p+1)
                # ax.annotate(p+1, xy=(p * 3 + 1, -57), xycoords=('data', 'axes points'), ha='center', va='bottom')
                # ax.annotate(p + 1, xy=(-57, p * 3 + 1), xycoords=('axes points', 'data'), ha='left', va='center')
            else:
                text = '    '
            ax.annotate(text, xy=(p*3+1+xi*2.3/3, -17), xytext=(p*3+1, -20), xycoords=('data', 'axes points'),
                        textcoords=('data', 'offset points'),
                        ha='center', va='bottom', arrowprops={'width': 0.2, 'headwidth': 3, 'headlength': 2}, in_layout=True)
            ax.annotate(text, xy=(-17, p * 3 + 1 + xi * 2.3 / 3), xytext=(-25, p * 3 + 1), xycoords=('axes points', 'data'),
                        textcoords=('offset points', 'data'),
                        ha='left', va='center', arrowprops={'width': 0.2, 'headwidth': 3, 'headlength': 2}, in_layout=True)


    ax.set_xlabel('Mutation 1', ha='center', labelpad=25)
    ax.set_ylabel('Mutation 2', labelpad=30)
    # ax.xaxis.set_label_coords(0.5, -0.12)
    # ax.yaxis.set_label_coords(-0.13, 0.5)

    if da_annotation is not None:
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                value = data_annotation[i, j]
                if ~np.isnan(value):
                    text = ax.text(j, i, value.astype(int),
                                   ha="center", va="center", color="k")

    if save and save_path is not None:
        save_path = Path(save_path)
        fig.savefig(save_path / (sequence + '_double_mutations_' + da.name + '.pdf'))
        fig.savefig(save_path / (sequence + '_double_mutations_' + da.name + '.png'))

    return fig, ax



def plot_sequencing_match(match, write_path, title, filename, unit = 'um', MiSeq_pixels_to_um = None, Fluo_pixels_to_um = None, save=True):
    # TODO: Update for new tile and sequencing matches
    source = match.source
    source_in_destination = match.transform_coordinates(source)
    destination = match.destination
    destination_in_source = match.transform_coordinates(destination, inverse=True)

    source_vertices = match.source_vertices
    destination_vertices = match.transform_coordinates(match.source_vertices)

    if unit == 'um':
        source = Fluo_pixels_to_um(source)
        source_in_destination = MiSeq_pixels_to_um(source_in_destination)
        destination = MiSeq_pixels_to_um(destination)
        destination_in_source = Fluo_pixels_to_um(destination_in_source)

        source_vertices = Fluo_pixels_to_um(source_vertices)
        destination_vertices = MiSeq_pixels_to_um(destination_vertices)

    # fig = plt.figure(figsize = (8,4))
    # ax1, ax2 = fig.subplots(1, 2)

    fig, ax1 = plt.subplots(figsize = (8,4))
    fig.subplots_adjust(0.05,0.05,0.95,0.93)

    divider = make_axes_locatable(ax1)
    ax2 = divider.append_axes('right', size='60%', pad=0.5)
    #
    # from mpl_toolkits.axes_grid1 import ImageGrid
    #
    # fig = plt.figure(figsize = (8,4))
    # grid = ImageGrid(fig, 111, nrows_ncols=(1,2), axes_pad=0.1, add_all=True, label_mode='L')
    #
    # ax1 = grid[0]
    # ax2 = grid[1]

    ax1.scatter(source_in_destination[:,0],source_in_destination[:,1],c='#40A535',marker = 'x')
    ax1.scatter(destination[:,0],destination[:,1], marker = '.', facecolors = 'k', edgecolors='k')
    ax1.set_facecolor('white')
    ax1.set_aspect('equal')


    ax1.set_xlim([0, 30000])
    ax1.set_ylim([0, 30000])
    ax1.set_xlabel('x')
    ax1.set_ylabel('y')

    ax1.ticklabel_format(axis='both', style='sci', scilimits=(5,6), useOffset=None, useLocale=None, useMathText=True)

    if unit == 'um':
        start = 0
        end = 1001
        stepsize = 250

        ax1.set_xlim([0, 1000])
        ax1.set_ylim([0, 1000])
        ax1.set_xlabel('x (\u03BCm)')
        ax1.set_ylabel('y (\u03BCm)')
        ax1.xaxis.set_ticks(np.arange(start, end, stepsize))
        ax1.yaxis.set_ticks(np.arange(start, end, stepsize))

        ax1.ticklabel_format(axis='both', style='sci', scilimits=(0,4), useOffset=None, useLocale=None, useMathText=True)

    p = patches.Polygon(destination_vertices,
                        linewidth=2, edgecolor='#40A535', facecolor='none', fill='false'
                        )

    ax1.add_patch(p)


    #plt.tight_layout()

    # fig.savefig(write_path.joinpath(name + '.pdf'), bbox_inches='tight')
    # fig.savefig(write_path.joinpath(name + '.png'), bbox_inches='tight', dpi=1000)


    ax2.scatter(source[:,0],source[:,1],c='#40A535',marker = 'x')
    ax2.scatter(destination_in_source[:,0],destination_in_source[:,1], marker = '.', facecolors = 'k', edgecolors='k')
    ax2.set_facecolor('white')
    ax2.set_aspect('equal')

    image_size = np.array([np.min(source_vertices, axis=0),np.max(source_vertices, axis=0)])

    ax2.set_xlim([image_size[0,0], image_size[1,0]])
    ax2.set_ylim([image_size[0,1], image_size[1,1]])

    ax2.set_xlabel('x')
    ax2.set_ylabel('y')

    ax2.ticklabel_format(axis='both', style='sci', scilimits=(5,6), useOffset=None, useLocale=None, useMathText=True)

    if unit == 'um':
        # ax2.set_xlim([image_size[0, 0], image_size[1, 0]])
        # ax2.set_ylim([image_size[0, 1], image_size[1, 1]])
        ax2.set_xlabel('x (\u03BCm)')
        ax2.set_ylabel('y (\u03BCm)')

        ax2.ticklabel_format(axis='both', style='sci', scilimits=(0,4), useOffset=None, useLocale=None, useMathText=True)

    # if ax1pos is not None:
    #     ax1.set_position(ax1pos)
    #     ax2.set_position(ax2pos)

    for spine in ax2.spines.values():
        spine.set_edgecolor('#40A535')

    def connect_vertices_in_axis(verticesA, verticesB, axisA, axisB, **kwargs):
        for vertexA, vertexB in zip(verticesA, verticesB):
            con = ConnectionPatch(xyA=vertexA, xyB=vertexB, coordsA="data", coordsB="data",
                                  axesA=axisA, axesB=axisB, **kwargs)
            axisB.add_artist(con)

    connect_vertices_in_axis(source_vertices, destination_vertices, ax2, ax1)

    for artist in ax1.artists:
        artist.set_linestyle((0,(5,5)))
        artist.set_linewidth(0.5)
        artist.set_edgecolor('grey')

    fig.suptitle(title, fontsize='medium')

    plt.show()

    if save:
        n = filename.replace('\\', '_')
        fig.savefig(write_path.joinpath(n + '.pdf'), bbox_inches='tight')
        fig.savefig(write_path.joinpath(n + '.png'), bbox_inches='tight', dpi=250)

    return ax1, ax2



# Show all matched files in tiles
def plot_matched_files_in_tile(files, show_file_coordinates=False, show_file_vertices=True, unit='um', save=False):
    # TODO: Update for new tile and sequencing matches

    def MiSeq_pixels_to_um(pixels):
        return 958 / 2800 * (pixels - 1000) / 10

    for tile in files[0].experiment.sequencing_data_for_mapping.tiles:
        files_on_tile = [file for file in files if file.sequencing_match.tile == tile.number]
        print(len(files_on_tile))

        tile_coordinates = tile.coordinates

        if unit == 'um':
            tile_coordinates = MiSeq_pixels_to_um(tile_coordinates)

        figure = plt.figure()
        axis = figure.gca()

        axis.scatter(tile_coordinates[:,0], tile_coordinates[:,1], marker = '.', facecolors = 'k', edgecolors='k')
        axis.set_facecolor('white')
        axis.set_aspect('equal')

        axis.set_xlim([0, 30000])
        axis.set_ylim([0, 30000])
        axis.set_xlabel('x (FASTQ)')
        axis.set_ylabel('y (FASTQ)')

        axis.ticklabel_format(axis='both', style='sci', scilimits=(5,6), useOffset=None, useLocale=None, useMathText=True)

        for file in files_on_tile:
            if show_file_coordinates:
                coordinates = file.sequencing_match.source_to_destination
                if unit == 'um':
                    coordinates = MiSeq_pixels_to_um(coordinates)
                axis.scatter(coordinates[:, 0], coordinates[:, 1], c='#40A535', marker='x')

            # figure.gca().scatter(vertices[:, 0], vertices[:, 1], c='g')
            if show_file_vertices:
                vertices = file.sequencing_match.transform_coordinates(file.movie.channels[1].vertices)
                if unit == 'um':
                    vertices = MiSeq_pixels_to_um(vertices)
                p = patches.Polygon(vertices,
                                    linewidth=2, edgecolor='#40A535', facecolor='none', fill='false'
                                    )

                axis.add_patch(p)

        axis.set_title(f'Tile {tile.name}')

        if unit == 'um':
            start = 0
            end = 1001
            stepsize = 250

            axis.set_xlim([0, 1000])
            axis.set_ylim([0, 1000])
            axis.set_xlabel('x (\u03BCm)')
            axis.set_ylabel('y (\u03BCm)')
            axis.xaxis.set_ticks(np.arange(start, end, stepsize))
            axis.yaxis.set_ticks(np.arange(start, end, stepsize))

            axis.ticklabel_format(axis='both', style='sci', scilimits=(0, 4), useOffset=None, useLocale=None,
                                 useMathText=True)

        if save:
            figure.tight_layout()
            figure.savefig(f'Matched_files_in_tile_{tile.name}.png', bbox_inches='tight', dpi=250)

def plot_cluster_locations_per_tile(dataset, number_of_tiles=19, number_of_surfaces=2, save_filepath=None):
    # TODO: Update for new tile and sequencing matches
    # df should contain Tile number, x and y

    number_of_rows = number_of_surfaces * math.ceil(number_of_tiles / 10)
    number_of_columns = np.min([10, number_of_tiles])
    figure, axes = plt.subplots(number_of_rows, number_of_columns, sharex=True, sharey=True,
                                figsize=(number_of_columns*2, number_of_rows*2))

    for ax in axes.flatten():
        row_index, column_index = np.where(axes == ax)
        row_index = row_index[0]
        column_index = column_index[0]

        side_index = row_index // 2
        tile_index = column_index + row_index % 2 * 10
        tile_number = (side_index+1)*1000+100+tile_index+1

        # print(side_index, tile_index, row_index, column_index, tile_number)

        if tile_index+1 > number_of_tiles:
            ax.set_visible(False)
            continue

        dataset_tile = dataset[{'sequence': dataset.tile == tile_number}]

        dataset_tile.to_dataframe().plot.scatter(x='x', y='y', ax=ax, marker='.', s=7)
        ax.set_title(tile_number)
        ax.set_aspect('equal')
        ax.set_xlabel('x (sequencer)')
        ax.set_ylabel('y (sequencer)')


    figure.tight_layout()
    if save_filepath:
        figure.savefig(save_filepath, dpi=300)


from PySide2.QtWidgets import QApplication

from papylio.gui.main import MainWindow
import sys

from multiprocessing import Process, freeze_support


def start_gui():
    freeze_support()

    app = QApplication(sys.argv)

    window = MainWindow()
    window.show()

    app.exec_()


if __name__ == '__main__':
    start_gui()


# import numpy as np #scientific computing with Python
# import matplotlib.pyplot as plt #Provides a MATLAB-like plotting framework
# import pandas as pd
# import xarray as xr
# from pathlib import Path
# from papylio.analysis.autoThreshold import stepfinder
# # from papylio.plugin_manager import PluginManager
# # from papylio.plugin_manager import PluginMetaClass
# from papylio.plugin_manager import plugins
# from papylio.trace_extraction import make_gaussian
# import copy
#
# class Molecules:
#     def load(filepath):
#         return Molecules().load(filepath)
#
#     def sum(list_of_molecules):
#         new = Molecules()
#         new.dataset = xr.concat([m.dataset for m in list_of_molecules], dim='molecule')
#         return new
#
#     def __init__(self, dataset=None):#, file, number_of_molecules):
#         # for value in molecules:
#         #     if not isinstance(value, Molecule):
#         #         raise TypeError('MoleculeList can only contain Molecule objects')
#         # self.molecules = molecules
#         # self.name = name
#
#         if dataset is None:
#             self.dataset = xr.Dataset()
#             self.init_dataset(reset=True)
#
#
#     # @property
#     # def dataset(self):
#     #     return self._dataset
#     #
#     # @dataset.setter
#     # def dataset(self, dataset):
#     #     if not self._dataset:
#     #         self._dataset = dataset
#     #         if 'selected' not in self._dataset.keys():
#     #             self.dataset['selected'] = xr.DataArray(False, coords=[dataset.molecule])
#     #         elif 'selected' not in self._dataset.keys():
#     #             self.dataset['x'] = xr.DataArray(False, coords=[dataset.molecule])
#     #             self.dataset['y'] = xr.DataArray(False, coords=[dataset.molecule])
#     #     else:
#     #         self._dataset = dataset
#
#     def init_dataset(self, molecule_multiindex=None, reset=False):
#         if not reset and len(self.dataset.molecule) > 0:
#             return
#         if molecule_multiindex is None:
#             molecule_multiindex = pd.MultiIndex.from_tuples([], names=['molecule_in_file', 'file'])
#         # self.dataset = xr.Dataset(
#         #     {
#         #         'selected':     ('molecule', xr.DataArray(False, coords=(molecule_multiindex,)),
#         #         'x':            ('molecule', xr.DataArray(np.nan, coords=[molecule_multiindex])),
#         #         'y':            ('molecule', xr.DataArray(np.nan, coords=[molecule_multiindex])),
#         #         'background':   ('molecule', xr.DataArray(np.nan, coords=[molecule_multiindex]))
#         #         'traces':       (('molecule'))
#         #     },
#         #     coords=
#         #     {
#         #         'molecule': ('molecule', molecule_multiindex), # pd.MultiIndex.from_tuples([], names=['molecule_in_file', 'file'])),
#         #         'frame': ('frame', np.array([], dtype=int)),
#         #         'channel': ('channel', np.array([], dtype=int))
#         #     }
#         # )
#
#         self.dataset = xr.Dataset(
#             {
#                 'selected':     ('molecule', xr.DataArray(False, coords=[molecule_multiindex])),
#                 'x':            (('molecule', 'channel'), xr.DataArray(np.nan, coords=[molecule_multiindex,[]])),
#                 'y':            (('molecule', 'channel'), xr.DataArray(np.nan, coords=[molecule_multiindex, []])),
#                 'background':   (('molecule','channel'), xr.DataArray(np.nan, coords=[molecule_multiindex, []])),
#                 'traces':       (('molecule', 'channel', 'frame'), xr.DataArray(np.nan, coords=[molecule_multiindex,[],[]]))
#             },
#             coords=
#             {
#                 'molecule': ('molecule', molecule_multiindex), # pd.MultiIndex.from_tuples([], names=['molecule_in_file', 'file'])),
#                 'frame': ('frame', np.array([], dtype=int)),
#                 'channel': ('channel', np.array([], dtype=int))
#             }
#         )
#
#         # self.empty_parameter_index = pd.MultiIndex.from_arrays([['is_selected'],[]], names=['Parameter','Channel'])
#         # self.empty_traces_index = pd.MultiIndex.from_arrays([[],[]], names=['Frame','Channel'])
#         # self._traces = None
#         # self._parameters = None
#         # self.channels = [0,1]
#
#     # @property
#     # def traces(self):
#     #     return self._traces
#     #
#     # @traces.setter
#     # def traces(self, traces):
#     #     if self._traces is not None and (len(traces.columns) != len(self.molecule_index)):
#     #         raise ValueError('Number of molecules does not match current number of molecules, to proceed reset the object first ')
#     #     self._traces = traces
#     #     self.molecule_index = traces.columns
#     #
#     # @property
#     # def parameters(self):
#     #     return self._parameters
#     #
#     # @parameters.setter
#     # def parameters(self, parameters):
#     #     if self._parameters is not None and (len(parameters.columns) != len(self.molecule_index)):
#     #         raise ValueError('Number of molecules does not match current number of molecules, to proceed reset the object first ')
#     #     self._parameters = parameters
#     #     self.molecule_index = parameters.columns
#     #
#     # @property
#     # def molecule_index(self):
#     #     if self._traces is not None:
#     #         return self.traces.columns
#     #
#     # @molecule_index.setter
#     # def molecule_index(self, molecule_index):
#     #     if self._traces is None:
#     #         self._traces = pd.DataFrame(index=self.empty_traces_index, columns=molecule_index)
#     #     else:
#     #         self._traces.columns = molecule_index
#     #     if self._parameters is None:
#     #         self._parameters = pd.DataFrame(index=self.empty_parameter_index, columns=molecule_index)
#     #     else:
#     #         self._parameters.columns = self.molecule_index
#     #
#     # def reset(self):
#     #     self._traces = None
#     #     self._parameters = None
#
#     def __len__(self):
#         return len(self.dataset.molecule)
#
#     def __getitem__(self, item):
#         new = Molecules()
#         new.dataset = self.dataset.isel(molecule=item)
#         return new
#
#     def __add__(self, other):
#         new = Molecules()
#         new.dataset = xr.concat([self.dataset, other.dataset], dim='molecule')
#         return new
#
#     def __radd__(self, other):
#         if other == 0:
#             return self
#         else:
#             return self.__add__(other)
#
#     def __getattr__(self, name):
#         try:
#             return self.dataset[name]
#         except KeyError:
#             super().__getattribute__(name)
#
#     # def __setattr__(self, key, value):
#     #     if key in self.__dict__.keys():
#     #         super().__setattr__(key, value)
#     #     else:
#     #         self._parameters.loc[key,:] = value
#
#     # def add_parameters(self, added_parameters):
#     #     parameters = self.parameters.append(added_parameters)
#     #     self.parameters = parameters[~parameters.index.duplicated(keep='last')]
#     #
#     # def add_parameter_from_list(self, name, parameter_list, channel=''):
#     #     self.parameters.loc[(name, channel), :] = parameter_list
#
#     @property
#     def coordinates(self):
#         return self.dataset[['x', 'y']].to_array('dimension')
#
#     @coordinates.setter
#     def coordinates(self, coordinates):
#         self.dataset = self.dataset.merge(coordinates.to_dataset('dimension'))
#
#     # def select_with_coords(self, kwargs):
#     #     self.dataset[] == 'MapSeq'
#
#     def import_file(self, filepath):
#         filepath = Path(filepath)
#         if filepath.suffix == '.traces':
#             self.import_traces_file(filepath)
#         elif filepath.suffix == '.pks':
#             self.import_pks_file(filepath)
#         else:
#             raise FileNotFoundError(filepath)
#
#     def save(self, filepath):
#         pass
#         # filepath = Path(filepath)
#         # self.dataset.reset_index('molecule').to_netcdf(filepath.with_suffix('.nc'))
#
#     def load(self, filepath):
#         pass
#         # filepath = Path(filepath)
#         # loaded_dataset = xr.open_dataset(filepath.with_suffix('.nc')).set_index({'molecule': ('molecule_in_file','file')})
#         # self.init_dataset(loaded_dataset.molecule)
#         # self.dataset = loaded_dataset.combine_first(self.dataset)
#
#
#
#
#     # column_index = pd.MultiIndex.from_product([[traces_filepath.with_suffix('').name],
#     #                                     np.arange(number_of_molecules),
#     #                                     np.arange(number_of_channels)],
#     #                                    names=['File', 'Molecule', 'Channel'])
#     # index = pd.Index(data=np.arange(number_of_frames), name='Frame')
#     # return pd.DataFrame(traces.T, index=index, columns=column_index).stack('Channel')
#
#
#
# # test.reset_index('trace').set_index(trace=['file','molecule','channel'])
# # test3.reset_index('trace').reset_coords('file', drop=True).assign_coords({'file': ('trace',[1]*278)}).set_index(trace=['file','molecule','channel'])
# # number_of_channels = 2
# # number_of_molecules = a.shape[a.dims=='trace']
# # traces = traces.unstack('Channel')
#
#
#
#
# @plugins
# class Molecule:
#     pass
# #     slots = ('file', 'index', '_coordinates', 'intensity', '_background', 'is_selected', 'steps', 'kon_boolean')
# #
# #     def __init__(self, file):
# #         self.file = file
# #         self.index = None
# #         self._coordinates = None
# #         self.intensity = None
# #         self._background = None
# #
# #         self.is_selected = False
# #
# #         self.steps = None  #Defined in other classes as: pd.DataFrame(columns=['frame', 'trace', 'state', 'method','thres'])
# #         self.kon_boolean = None  # 3x3 matrix that is indicates whether the kon will be calculated from the beginning, in-between molecules or for the end only
# #         #self.bg_scale=np.sum(make_gaussian(self.file.experiment.configuration['find_coordinates']['coordinate_optimization']['coordinates_after_gaussian_fit']['gaussian_width']))
# #
# #     @property
# #     def coordinates(self):
# #         return self._coordinates
# #
# #     @property
# #     def background(self):
# #         return self._background
# #
# #     @coordinates.setter
# #     def coordinates(self, coordinates):
# #         self._coordinates = np.atleast_2d(coordinates)
# #
# #     def background(self, background):
# #         self.background=background # should be dependent on emission channel as well
# #
# #     @property  # this is just for the stepfinder to be called through Molecule. Maybe not needed
# #     def find_steps(self):
# #         return stepfinder
# #
# #     def I(self, emission, Ioff=0):
# #         return self.intensity[emission, :] - Ioff # - self.background[emission] * self.bg_scale #this number comes from sum(make_gaussian) in trace_extraction
# #
# #     def E(self, Imin=0, Iroff=0, Igoff=0, alpha=0):
# #         red = np.copy(self.I(1, Ioff=Iroff))
# #         green = self.I(0, Ioff=Igoff)
# #         np.putmask(green, green < 0, 0) # green < 0 is taken as 0
# #         np.putmask(red, red < Imin, 0)  # the mask makes all elements of acceptor that are below the Imin zero, for E caclulation
# #         E =  (red - alpha*green) / (green + red - alpha*green)
# #         E = np.nan_to_num(E)  # correct for divide with zero = None values
# #         return E
# #
# #     def plot(self, ylim=(0, 500), xlim=(), Ioff=[],  save=False, **fretkwargs):
# #         plt.style.use('seaborn-dark')
# #         plt.style.use('seaborn-colorblind')
# #         figure = plt.figure(f'{self.file.name}_mol_{self.index}', figsize=(7,4))
# #         if len(self.file.experiment.pairs) > 0:
# #             axis_I = figure.add_subplot(211)
# #         else:
# #             axis_I = figure.gca()
# #
# #         axis_I.set_ylabel('Intensity (a.u.)')
# #         axis_I.set_ylim(ylim[0], ylim[1])
# #         if xlim == ():
# #             axis_I.set_xlim(0, self.file.time.max()+1)
# #         else:
# #             axis_I.set_xlim(xlim[0], xlim[1])
# #
# #         axis_I.set_title(f'Molecule {self.index} /{len(self.file.molecules)}')
# #         if Ioff == []:
# #             Ioff = [0]*self.file.number_of_channels
# #         for i, channel in enumerate(self.file.experiment.channels):
# #             axis_I.plot(self.file.time, self.I(i, Ioff=Ioff[i]), channel)
# #
# #         if len(self.file.experiment.pairs) > 0:
# #             axis_E = figure.add_subplot(212, sharex=axis_I)
# #             axis_E.set_xlabel('Time (s)')
# #             axis_E.set_ylabel('FRET')
# #             axis_E.set_ylim(0,1.1)
# #             for i, pair in enumerate(self.file.experiment.pairs):
# #                 axis_E.plot(self.file.time, self.E(**fretkwargs), 'b')
# #
# #         plt.tight_layout()
# #         if save:
# #             plt.savefig(f'{self.file.relativeFilePath}_mol_{self.index}.eps', transparent=True)
# #             plt.savefig(f'{self.file.relativeFilePath}_mol_{self.index}.png', facecolor='white', dpi=300, transparent=True)
# #
#
#
#
#
#
# if __name__ == '__main__':
#     filepath = r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy\test'
#     # traces_filepath = r'P:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy\test'
#     filepath = Path(filepath)
#
#     test = Molecules()
#     test.import_traces_file(filepath.with_suffix('.traces'))
#     test.import_pks_file(filepath.with_suffix('.pks'))
#
#     # pks_filepath = Path(traces_filepath).with_suffix('.pks')
#     # test.import_pks_file(pks_filepath, 2)
#     # test.export_pks_file(pks_filepath.with_name('test2.pks'))
#     #
#     #
#     # test[5:10]
#     #
#     # test[5]+test[10]
#
#
#     test.export_traces_file(filepath.with_name('test2.traces'))
#     test.export_pks_file(filepath.with_name('test2.pks'))


import numpy as np
import scipy.optimize
import scipy.ndimage
import tqdm

from papylio.movie.shading_correction import get_photobleach

def determine_temporal_background_correction(frames, method, flatfield=None, darkfield=None):
    if method == 'BaSiC':
        channel_dimensions = frames.shape[-2:][::-1]  # size should be given in (x,y) for get_photobleach
        size = (channel_dimensions / np.max(channel_dimensions) * 256).astype(int)
        correction = get_photobleach(frames, flatfield, darkfield, size=size).flatten()
    elif method == 'BaSiC_crop':
        # Crop the image instead of resizing, may be better for single-molecules. However, may give higher noise.
        raise NotImplementedError('')
    else:
        correction = np.array([determine_single_value_background_correction(frame, method, flatfield, darkfield)
                               for frame in tqdm.tqdm(frames)])

    return correction


def determine_single_value_background_correction(frame, method, flatfield=None, darkfield=None):
    frame_corrected = frame
    if darkfield is not None:
        frame_corrected = frame - darkfield
    if flatfield is not None:
        frame_corrected = frame_corrected / flatfield

    if method == 'BaSiC':
        channel_dimensions = frame.shape[-2:][::-1]  # size should be given in (x,y) for get_photobleach
        size = (channel_dimensions / np.max(channel_dimensions) * 256).astype(int)
        correction = get_photobleach(frame, flatfield, darkfield, size=size).flatten()
    elif method == 'BaSiC_crop':
        # Crop the image instead of resizing, may be better for single-molecules. However, may give higher noise.
        raise NotImplementedError('')
    elif method == 'mean':
        correction = frame_corrected.mean()
    elif method == 'median':
        correction = np.median(frame_corrected)
    elif method == 'fit_background_peak':
        correction = gaussian_maximum_fit(frame_corrected)
    elif 'filter' in method:
        correction = determine_spatial_background_correction(frame_corrected, method).mean()  # TODO: take mean only over x, y not over channel
    else:
        raise ValueError(f'Method {method} not found')

    return correction


def determine_spatial_background_correction(frame, method, flatfield=None, darkfield=None, **kwargs):
    # TODO: improve edge between donor&acceptor channel
    if method == 'gaussian_filter':
        scipy_filter = scipy.ndimage.gaussian_filter
        scipy_filter_kwargs = {'sigma': 0.5, 'mode': 'wrap'}
        # This comes down to taking the mean of the corrected image
    elif method == 'minimum_filter':
        scipy_filter = scipy.ndimage.minimum_filter
        scipy_filter_kwargs = {'size': 15, 'mode': 'wrap'}
    elif method == 'median_filter':
        scipy_filter = scipy.ndimage.median_filter
        scipy_filter_kwargs = {'size': 30, 'mode': 'wrap'}
    else:
        raise ValueError(f'Method {method} not found')

    if darkfield is not None:
        frame = frame - darkfield
    if flatfield is not None:
        frame = frame / flatfield

    scipy_filter_kwargs.update(kwargs)

    correction = scipy_filter(frame, **scipy_filter_kwargs)

    return correction

def gauss_function(x, a, x0, sigma):
    return a * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))

def gaussian_maximum_fit(frame, width_around_peak_fitted=200):
    # count, edges = np.histogram(frame.flatten(), bins=width_around_peak_fitted)
    # max_bin_center = edges[count.argmax():count.argmax()+2].mean()

    frame_min = np.floor(frame.min()).astype(int)
    frame_max = int(np.quantile(frame, 0.999))  # np.ceil(frame.max()).astype(int)

    frame_flattened = frame.flatten()

    # factor = 10
    # number_of_bins = int((frame_max - frame_min + 1)/factor)
    number_of_bins = 500
    factor = (frame_max - frame_min + 1)/number_of_bins

    count, edges = np.histogram(frame_flattened, bins=number_of_bins, range=(frame_min - 0.5, frame_max + 0.5))

    bincenters = (edges[:-1] + edges[1:]) / 2
    max_bin_center = bincenters[count.argmax()]

    width = width_around_peak_fitted

    x_peak = None
    while x_peak is None:
        selection = np.vstack(
            [max_bin_center - width / 2 < bincenters, bincenters < max_bin_center + width / 2]).all(axis=0)
        x = bincenters[selection]
        y = count[selection]

        std = np.std(frame_flattened[(frame_flattened > x.min()) & (frame_flattened < x.max())])
        try:
            popt, pcov = scipy.optimize.curve_fit(gauss_function, x, y, p0=[count.max(), max_bin_center, std], maxfev=2000)
            x_peak = popt[1]
        except RuntimeError:
            # x_peak = gaussian_maximum_fit(frame, width_around_peak_fitted*2)
            width *= 2

    # plt.figure();
    # plt.hist(frame_flattened, bins=number_of_bins, range=(frame_min - 0.5, frame_max + 0.5))
    # plt.plot(x, gauss_function(x, *popt))
    return x_peak






 #     if method == 'rollingball':
    #         background = rollingball(image, self.width_pixels / 10)[1]  # this one is not used in pick_spots_akaze
    #         image_correct = image - background
    #         image_correct[image_correct < 0] = 0
    #         threshold = get_threshold(image_correct)
    #         return remove_background(image_correct, threshold)
    #     elif method == 'per_channel':  # maybe there is a better name
    #         sh = np.shape(image)
    #         threshold_donor = get_threshold(self.get_channel(image, 'donor'))
    #         threshold_acceptor = get_threshold(self.get_channel(image, 'acceptor'))
    #         background = np.zeros(np.shape(image))
    #         background[:, 0:sh[0] // 2] = threshold_donor
    #         background[:, sh[0] // 2:] = threshold_acceptor
    #         return remove_background(image, background)
    #
    #     # note: optionally a fixed threshold can be set, like with IDL
    #     # note 2: do we need a different threshold for donor and acceptor?






































# -*- coding: utf-8 -*-
"""
Created on Mon Apr 15 15:16:25 2019

@author: mwdocter
this code takes an images, calculates the rolling_ball background, and subtracts it
"""
#http://imagejdocu.tudor.lu/doku.php?id=gui:process:subtract_background
#Based on the a „rolling ball“ algorithm described in Stanley Sternberg's article, „Biomedical Image Processing“, IEEE Computer, January 1983. 
import scipy.ndimage as scim
import skimage
#from skimage.morphology import ball
import matplotlib.pyplot as plt

def rollingball(*args): # Matlab[im_out,im_bg]=rollingball(im_in,size_ball,im_bg)
    varargin = args
    im_in=varargin[0]
    if len(varargin)==1:
        size_ball=30
    else:
        size_ball=varargin[1]
    if len(varargin)<3:
        # from https://stackoverflow.com/questions/29320954/rolling-ball-background-subtraction-algorithm-for-opencv
        # Create 3D ball structure
        s = skimage.morphology.ball(size_ball)
        # Take only the upper half of the ball
        h = int((s.shape[1] + 1) / 2)
        # Flat the 3D ball to a weighted 2D disc
        s = s[:h, :, :].sum(axis=0)
        # Rescale weights into 0-255
        s = (255 * (s - s.min())) / (s.max()- s.min())
        ss=s[2*h//4:2*3*h//4,2*h//4:2*3*h//4]
        ss = (255 * (ss - ss.min())) / (ss.max()- ss.min())
       
        #im_bg=scim.grey_closing(im_in,structure=ss)
        im_bg=skimage.morphology.opening(im_in,ss)
        #im_out = scim.white_tophat(im, structure=s)
    else:
        im_bg=varargin[2]
        
    im_out=im_in-im_bg #note match 3s dimension im_bg to im_in
    im_out[im_out<0]=0
    
    return im_out, im_bg
    # Use im-opening(im,ball) (i.e. white tophat transform) (see original publication)
    

# def get_threshold(image_stack, show=0):
#     ydata = (np.sort(image_stack.ravel()))
#     ydata_original = ydata
#     xdata = np.array(range(0, len(ydata)))
#     # scale the data to make x and y evenly important
#     ymaxALL = float(max(ydata))
#     xmaxALL = float(max(xdata))
#     ydata = ydata * xmaxALL / ymaxALL  # don't forget this is scaled
#
#     # fit a line through the lowest half of x
#     xd = xdata[:int(np.floor(len(xdata) / 2))]
#     yd = ydata[:int(np.floor(len(xdata) / 2))]
#     p_start = np.polyfit(xd, yd, 1)
#
#     # fit a line through the upper half of y
#     ymax = max(ydata)
#     yhalf = ymax / 2
#     x2 = np.argwhere(abs(ydata - yhalf) == min(abs(ydata - yhalf)))
#     x2 = int(x2[0])
#     xd = xdata[x2:]
#     yd = ydata[x2:]
#     p_end = np.polyfit(xd, yd, 1)
#
#     # find the crossing of these lines
#     # a1*x+b1=a2*x+b2
#     # (a1-a2)*x=b2-b1
#     # x=(b2-b1)/(a1-a2)
#     x_cross = int((p_end[1] - p_start[1]) / (p_start[0] - p_end[0]))
#     y_cross = int(np.polyval(p_start, x_cross))
#
#     # add polyfits to the plot
#     y_fit_start = np.polyval(p_start, xdata[:x_cross])
#     x_fit_end = xdata[x_cross:]  # start to draw from crossing y=0
#     y_fit_end = np.polyval(p_end, x_fit_end)
#     # now find the closest distance from cross to actual data. x and y should be simarly scaled
#     xx = xdata - x_cross
#     xx = [float(ii) for ii in xx]
#     yy = ydata - y_cross
#     yy = [float(ii) for ii in yy]
#     rr = (np.array(xx) ** 2 + np.array(yy) ** 2) ** 0.5  # int32 is not large enough
#     x_found = np.argwhere(min(rr) == rr)
#     x_found = x_found[0, 0]
#     if show:
#         plt.figure()
#         fig2 = plt.subplot(1, 2, 2)
#         fig2.plot(xdata, rr * ymaxALL / xmaxALL)
#         # fig2.title("{:s}".format(x_found))
#
#         fig1 = plt.subplot(1, 2, 1)
#         fig1.plot(xdata, ydata * ymaxALL / xmaxALL, 'b')
#         fig1.plot(xdata[:x_cross], y_fit_start[:x_cross] * ymaxALL / xmaxALL, 'g')
#         fig1.plot(x_fit_end, y_fit_end * ymaxALL / xmaxALL, 'r')
#         fig1.plot(x_cross, y_cross * ymaxALL / xmaxALL, 'kx')
#
#         fig1.plot(x_found, ydata[x_found] * ymaxALL / xmaxALL, 'mo')
#
#         fig1.plot(xdata[:-1], ydata_original[1:] - ydata_original[:-1])
#         plt.show()
#
#     thr = ydata[x_found] * ymaxALL / xmaxALL
#     im_uit = image_stack - thr.astype(type(image_stack[0, 0]))
#     im_uit[im_uit < 0] = 0
#     return thr
#
#
# def remove_background(image_stack, thr, show=0):
#     im_uit = image_stack - thr.astype(type(image_stack[0, 0]))
#     im_uit[im_uit < 0] = 0
#     return im_uit
#
#


import PIL.ImageFilter
import numpy as np
import xarray as xr
from tqdm import tqdm
import dask_image.ndfilters
import scipy.ndimage

def make_gaussian_mask(size, offsets, sigma=1.291):
    # TODO: Explain calculation in docstring
    # It is to keep the photon number the same after applying the mask.
    # If there is a PSF of N photons, which is nothing but a 2D Gauss function with given sigma and amplitude,
    # the sum of the pixel is N. The idea is that the pixel sum should be the same after applying the mask.
    # The normalization factor is calculated to compensate the amplitude of 2D Gaussian after applying the mask.
    # The normalization factor should be different for different PSF size (i.e. different magnification or setup).
    # So N = sum(mask * (psf_single_photon*N)), and so sum(mask*psf_single_photon)
    # Both the mask and the psf are 2d Gaussians
    import xarray as xr
    roi = xr.DataArray(np.mgrid[0:size,0:size]-size//2, dims=('dimension','y','x'), coords={'dimension': ['y', 'x']})
    masks = np.exp(-((roi - offsets) ** 2).sum('dimension') / sigma**2 / 2).transpose('molecule','channel','y','x')
    psfs_single_photon = masks/masks.sum(dim=('x', 'y'))
    norm_factors = (masks*psfs_single_photon).sum(dim=('x','y'))
    masks = masks/norm_factors
    return masks

def extract_traces(movie, coordinates, background=None, mask_size=1.291, neighbourhood_size=11, correct_illumination=False):
    # go through all images, extract donor and acceptor signal
    # TODO: Process the movie in chunks
    # TODO: Make sure the corretions are not reloaded for each chunk,
    #  for example by loading them once at the with statement or by keeping recent variables in the cache/memmory

    coordinates['dimension'] = coordinates.dimension.astype('U')
    with movie:
        movie.read_header()

        if movie.number_of_frames > 500:
            all_frames_in_memory = False
        else:
            all_frames_in_memory = True

        intensity = xr.DataArray(np.empty((len(coordinates.molecule), len(coordinates.channel), movie.number_of_frames)),
                                 dims=['molecule', 'channel', 'frame'],
                                 coords=coordinates.drop('dimension').coords, name='intensity')

        # background_correction = xr.DataArray(np.empty((len(coordinates.molecule), len(coordinates.channel), movie.number_of_frames)),
        #                          dims=['molecule', 'channel', 'frame'],
        #                          coords=coordinates.drop('dimension').coords, name='background_correction')

        # channel_offsets = xr.DataArray(np.vstack([channel.origin for channel in movie.channels]),
        #                                dims=('channel', 'dimension'),
        #                                coords={'channel': [channel.index for channel in movie.channels],
        #                                        'dimension': ['x', 'y']}) # TODO: Move to Movie
        # coordinates = coordinates - channel_offsets

        # if background is None:
        #     background = xr.DataArray(dims=['molecule','channel'], coords={'molecule': coordinates.molecule, 'channel': coordinates.channel})

        offsets = coordinates % 1
        twoD_gaussians = make_gaussian_mask(size=neighbourhood_size, offsets=offsets, sigma=mask_size)

        coordinates_floored = (coordinates // 1).astype(int)

        roi_indices_general = xr.DataArray(np.mgrid[:neighbourhood_size, :neighbourhood_size] - neighbourhood_size // 2,
                                           dims=('dimension', 'y', 'x'),
                                           coords={'dimension': ['y', 'x']})#.transpose()

        roi_indices = coordinates_floored + roi_indices_general

        # if correct_illumination:
        #     illumination_correction = IlluminationCorrection(movie.number_of_frames,
        #                                                      filter_function=scipy.ndimage.minimum_filter,
        #                                                      size=15, mode='wrap')

        # background_per_frame = background.sel(illumination=movie.illumination)
        # background_correction[:] = weighed_background(background_per_frame, twoD_gaussians).transpose((1,2,0))

        if all_frames_in_memory:
            frames = movie.read_frames(xarray=False, flatten_channels=True)#.astype('uint16')

        oneD_indices = (roi_indices.sel(dimension='y')*movie.width+roi_indices.sel(dimension='x')).stack(peak=('molecule','channel')).stack(i=('y','x'))
        for frame_index in tqdm(range(movie.number_of_frames), desc=movie.name, leave=True):  # self.number_of_frames also works for pm, len(self.movie_file_object.filelist) not
            # print(frame_number)
            # if frame_number % 13 == 0:
            #     sys.stdout.write(f'\r   Frame {frame_number} of {movie.number_of_frames}')

            # image = movie.read_frame(frame_index, xarray=False, flatten_channels=True).astype('uint16')
            if all_frames_in_memory:
                image = frames[frame_index]
            else:
                image = movie.read_frame(frame_index, xarray=False, flatten_channels=True)#.astype('uint16')
            frame = xr.DataArray(image, dims=('y','x'))

            # TODO: Proper background subtraction

            # if correct_illumination:
            #     illumination_correction.add_frame(frame_index, frame)
            #     # TODO: Determine how illumination correction is dependent on background

            # if 'illumination' in background.dims:
            #     # TODO: Make this work properly and rename illumination in Movie
            #     # Do background subtraction on entire frame instead???
            #     frame_background = background.sel(illumination=movie.illumination.sel(frame=frame_index))
            #     # frame_background = background[frame_number % number_illumination]
            # else:
            #     frame_background = background

            #intensity[:, :, frame_index] = extract_intensity_from_frame(frame, background, roi_indices, twoD_gaussians)
            # intensity[:, :, frame_index] = extract_intensity_from_frame(frame, frame_background, oneD_indices, twoD_gaussians)
            intensity[:, :, frame_index] = extract_intensity_from_frame(frame, oneD_indices, twoD_gaussians)


        # sys.stdout.write(f'\r   Frame {frame_number+1} of {movie.number_of_frames}\n')
        dataset = intensity.to_dataset()
        # dataset['intensity_raw'] = dataset.intensity.copy()

        # if correct_illumination:
        #     dataset['illumination_correction'] = illumination_correction.illumination_correction
        #     dataset['intensity'] *= dataset['illumination_correction']

        # dataset['background_correction'] = background_correction
        # dataset['intensity'] -= dataset['background_correction'].sel(frame=0, drop=True)

    return dataset

# def extract_intensity_from_frame(frame, background, roi_indices, twoD_gaussians):
#     intensities = frame.sel(x=roi_indices.sel(dimension='x'), y=roi_indices.sel(dimension='y'))
#     intensities = intensities - background
#     weighted_intensities = intensities * twoD_gaussians
#     intensity_in_frame = weighted_intensities.sum(dim=('x', 'y'))
#     return intensity_in_frame

# A ufunc is probably better here
# def extract_intensity_from_frame(frame, background, roi_indices, twoD_gaussians):
#     intensities = frame.values[roi_indices.values[:,:,1,:,:], roi_indices.values[:,:,0,:,:]]
#     intensities = intensities - background.values[:,:,None,None]
#     weighted_intensities = intensities * twoD_gaussians.values
#     intensity_in_frame = weighted_intensities.sum(axis=(2,3))
#     return intensity_in_frame

def extract_intensity_from_frame(frame, oneD_indices, twoD_gaussians):  # extract traces
    intensities = frame.values.take(oneD_indices.values).reshape(twoD_gaussians.shape)
    # intensities = intensities - background.values[:,:,None,None]
    weighted_intensities = intensities * twoD_gaussians.values
    intensity_in_frame = weighted_intensities.sum(axis=(2,3))
    return intensity_in_frame

def weighed_background(background, twoD_gaussians):
    weighed_background_intensity = background.values[:, :, :, None, None] * twoD_gaussians.values[None, :, :, :, :]
    return weighed_background_intensity.sum(axis=(3, 4))

class IlluminationCorrection:
    # In time
    def __init__(self, number_of_frames, filter_function=scipy.ndimage.minimum_filter, **kwargs):
        self.filter_function = filter_function
        self.filter_kwargs = kwargs
        self._illumination_correction = np.empty(number_of_frames)

    def add_frame(self, index, frame):
        filtered_frame = self.filter_function(np.array(frame), **self.filter_kwargs)
        self._illumination_correction[index] = filtered_frame.sum()

    @property
    def illumination_correction(self):
        correction = self._illumination_correction.max() / self._illumination_correction
        return xr.DataArray(correction, dims=('frame',), name='illumination_correction')



# def illumination_intensity_from_frames(frames=None, filter_neighbourhood_size=15):
#     if frames is None:
#         frames = self.movie.read_frames_raw()
#
#     if not frames.chunks:
#         filtered_images = minimum_filter(frames, size=(1, 1, filter_neighbourhood_size, filter_neighbourhood_size))
#     else:
#         filtered_images = dask_image.ndfilters.minimum_filter(frames.data, size=(1, 1, filter_neighbourhood_size, filter_neighbourhood_size))
#
#     filtered_images = xr.DataArray(filtered_images, coords=frames.coords, name='illumination_correction')
#     illumination_intensity = filtered_images.sum(dim=('x','y'))
#     illumination_correction = (illumination_intensity.max(dim='frame') / illumination_intensity).T
#     # illumination_correction = illumination_correction.reset_index('frame', drop=True)
#     illumination_correction.to_netcdf(self.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 12 01:02:04 2019

@author: iason
"""
import wx

class Panel(wx.Frame):

    def __init__(self, parent=None, title='Plot and Fit dwelltime distributions'):
        wx.Frame.__init__(self, parent, title=title)

        self.InitUI()
        self.Centre()

    def InitUI(self):

        panel = wx.Panel(self)
        #  Top sizer which places the different boxes vertically
        topsizer = wx.BoxSizer(wx.VERTICAL)

        #  SizerFlags object for commonly used flags
        flagsExpand = wx.SizerFlags(0)
        flagsExpand.Expand().Border(wx.ALL, 5)

        flagsCenter = wx.SizerFlags(0)
        flagsCenter.Align(wx.CENTER).Border(wx.ALL, 5)

        flagsLeft = wx.SizerFlags(0)
        flagsLeft.Align(wx.LEFT).Border(wx.ALL, 5)

        # Grid sizer to hold dataset selection radiobutton and naming option
        sboxData = wx.StaticBox(panel, label="Data")
        sboxDataSizer = wx.StaticBoxSizer(sboxData, wx.VERTICAL)

        gridSizerDataSelect = wx.GridBagSizer(5, 5)
        self.radioDataSeparate = wx.RadioButton(sboxData,
                                                label='Analyze separately',
                                                style = wx.RB_GROUP)

        self.radioDataSeparate.Bind(wx.EVT_RADIOBUTTON, self.onRadioSeparatePress)

        self.radioDataCombine = wx.RadioButton(sboxData,
                                                label='Combine selected')

        self.radioDataCombine.Bind(wx.EVT_RADIOBUTTON, self.onRadioCombinePress)
        self.entryDataName = wx.TextCtrl(sboxData, size=(80,20))

        self.entryDataName.Disable()
        gridSizerDataSelect.Add(self.radioDataSeparate, pos=(0, 0), flag=wx.LEFT,
                                border=5)
        gridSizerDataSelect.Add(self.radioDataCombine, pos=(1, 0), flag=wx.LEFT,
                                border=5)
        gridSizerDataSelect.Add(wx.StaticText(sboxData, label='Dataset Name:'),
                                pos=(0, 1), flag=wx.LEFT, border=5)
        gridSizerDataSelect.Add(self.entryDataName, pos=(1, 1), flag=wx.LEFT,
                                border=5)

        sboxDataSizer.Add(gridSizerDataSelect, 1, wx.EXPAND)



        # Static box for input of distribution plot and fit options
        sboxConfig = wx.StaticBox(panel, label="Configuration")
        sboxConfigSizer = wx.StaticBoxSizer(sboxConfig, wx.VERTICAL)

        # Box to hold the distribution type choice
        sBoxSizerSelect = wx.BoxSizer(wx.HORIZONTAL)
        sBoxSizerSelect.Add(wx.StaticText(sboxConfig, label="Distribution: ",
                                          style=wx.ALIGN_LEFT), flagsCenter)



        self.comboDist = wx.ComboBox(sboxConfig, value='offtime',
                                        choices=['offtime', 'ontime'],
                                        style = wx.ALIGN_CENTRE)
        # Bind events to combobox. When the box is clicked
        self.comboDist.Bind(wx.EVT_COMBOBOX, self.DistConfigLoad)
        self.comboDist.Bind(wx.EVT_COMBOBOX_DROPDOWN, self.DistConfigSave)
        self.configuration = {key: {} for key in self.comboDist.GetItems()}
        # configparams = ['trace', 'side', 'min', 'max', 'scale', 'PlotType',
        #                 'binsize', 'FitBool', 'TmaxBool', 'model', 'Nfits']

        sBoxSizerSelect.Add(self.comboDist, flagsCenter)

        # GridBag Sizer to hold the parameter selection buttons
        gridSizer = wx.GridBagSizer(5, 5)
        # 1st column, choose the trace type(s)
        gridSizer.Add(wx.StaticText(sboxConfig, label='Trace:'),  pos=(0, 0),
                      flag=wx.LEFT, border=5)
        # 2nd column, choose the trace type(s)
        gridSizer.Add(wx.StaticText(sboxConfig, label='Side:'),  pos=(0, 1),
                      flag=wx.LEFT, border=5)

        traces = ['red', 'green', 'total', 'FRET']
        self.chbTraces = [wx.CheckBox(sboxConfig, label=l) for l in traces]
        sides = ['left', 'middle', 'right']
        self.chbSides = [wx.CheckBox(sboxConfig, label=l) for l in sides]

        for i, chb in enumerate(self.chbTraces):
            gridSizer.Add(chb, pos=(i+1,0), flag=wx.LEFT, border=5)

        for i, chb in enumerate(self.chbSides):
            gridSizer.Add(chb, pos=(i+1,1), flag=wx.LEFT, border=5)

        # Entry Boxes for min and max dwelltimes
        boxSizerMin = wx.BoxSizer(wx.HORIZONTAL)
        boxSizerMin.Add(wx.StaticText(sboxConfig, label='Min:'), flagsLeft)
        self.entryMinDwell = wx.TextCtrl(sboxConfig, value="0", size=(40,20))
        boxSizerMin.Add(self.entryMinDwell, flagsLeft)

        boxSizerMax = wx.BoxSizer(wx.HORIZONTAL)
        boxSizerMax.Add(wx.StaticText(sboxConfig, label='Max:'), flagsLeft)
        self.entryMaxDwell = wx.TextCtrl(sboxConfig, value="Max", size=(40,20))
        boxSizerMax.Add(self.entryMaxDwell, flagsLeft)

        gridSizer.Add(boxSizerMin,  pos=(5, 0), span=(1,2), flag=wx.LEFT, border=2)
        gridSizer.Add(boxSizerMax,  pos=(6, 0), span=(1,2), flag=wx.LEFT, border=2)

        # Controls for plot-related parameters input
        boxSizerPlotParams = wx.StaticBoxSizer(wx.HORIZONTAL, sboxConfig,
                                               label="Plot options")
        gridSizerPlotParams = wx.GridBagSizer(5, 5)
        # Add the scale choose combobox
        gridSizerPlotParams.Add(wx.StaticText(sboxConfig, label='Scale:'),
                                pos=(0, 0), flag=wx.LEFT, border=5)
        self.comboScale = wx.ComboBox(sboxConfig, value='Normal',
                                        choices=['Normal', 'Log', 'Log-Log'],
                                        style = wx.ALIGN_CENTRE)
        gridSizerPlotParams.Add(self.comboScale, pos=(0,1), flag=wx.LEFT,
                                 border=5)

        # Add the plot type choose combobox
        gridSizerPlotParams.Add(wx.StaticText(sboxConfig, label='Plot type:'),
                                pos=(1, 0), flag=wx.LEFT, border=5)
        self.comboPlotType = wx.ComboBox(sboxConfig, value='dots',
                                        choices=['dots', 'bars', 'line'],
                                        style = wx.ALIGN_CENTRE)
        gridSizerPlotParams.Add(self.comboPlotType, pos=(1,1), flag=wx.LEFT,
                                 border=5)

        # Add the plot binning entry box
        gridSizerPlotParams.Add(wx.StaticText(sboxConfig, label='Bin size:'),
                                pos=(0, 2), flag=wx.LEFT, border=5)
        self.entryBinSize = wx.TextCtrl(sboxConfig, value="auto", size=(45,20))
        gridSizerPlotParams.Add(self.entryBinSize, pos=(0,3), flag=wx.LEFT,
                                 border=5)
        boxSizerPlotParams.Add(gridSizerPlotParams, flagsLeft)

        # Controls for fit-related parameters input
        boxSizerFitParams = wx.StaticBoxSizer(wx.VERTICAL, sboxConfig,
                                               label="Fit options")

        gridSizerFitParams = wx.GridBagSizer(5, 5)


        self.chbFitBool = wx.CheckBox(sboxConfig, label='Fit')
        self.chbTmaxBool = wx.CheckBox(sboxConfig, label='Include > Tmax')
        self.chbBootBool = wx.CheckBox(sboxConfig, label='Bootstrap')
        self.comboFitModel = wx.ComboBox(sboxConfig, value='1Exp',
                                        choices=['1Exp', '2Exp',
                                                 '1Exp+2Exp', '3Exp'],
                                        style = wx.ALIGN_CENTRE)
        self.entryNfits = wx.TextCtrl(sboxConfig, value="10", size=(40,20))
        self.entryBoots = wx.TextCtrl(sboxConfig, value="100", size=(40,20))

        gridSizerFitParams.Add(self.chbFitBool, pos=(0,0),
                               flag=wx.LEFT, border=5)
        gridSizerFitParams.Add(self.chbTmaxBool, pos=(1,0),
                               flag=wx.LEFT, border=5)
        gridSizerFitParams.Add(self.chbBootBool, pos=(2,0),
                               flag=wx.LEFT, border=5)

        gridSizerFitParams.Add(wx.StaticText(sboxConfig, label='Model:'),
                                pos=(0, 1), flag=wx.LEFT, border=0)
        gridSizerFitParams.Add(wx.StaticText(sboxConfig, label='Best of:'),
                                pos=(1, 1), flag=wx.LEFT, border=0)
        gridSizerFitParams.Add(wx.StaticText(sboxConfig, label='Repeats:'),
                                pos=(2, 1), flag=wx.LEFT, border=0)
        gridSizerFitParams.Add(self.comboFitModel, pos=(0,3),
                               flag=wx.LEFT, border=0)
        gridSizerFitParams.Add(self.entryNfits, pos=(1,3),
                               flag=wx.LEFT, border=0)
        gridSizerFitParams.Add(self.entryBoots, pos=(2,3),
                               flag=wx.LEFT, border=0)

        boxSizerFitParams.Add(gridSizerFitParams)

        sboxConfigSizer.Add(sBoxSizerSelect, 1, wx.EXPAND)
        sboxConfigSizer.Add(gridSizer, 4, wx.EXPAND)
        sboxConfigSizer.Add(boxSizerPlotParams, 2, wx.EXPAND)
        sboxConfigSizer.Add(boxSizerFitParams, 2, wx.EXPAND)


        # Controls and button for save options
        self.gridSaveSizer = wx.GridBagSizer(5, 5)
        self.SavePlotsBool = wx.CheckBox(panel, label='Save Plots')
        self.SaveFitsBool = wx.CheckBox(panel, label='Save Fits')
        self.SaveButton = wx.Button(panel, label='Save')


        self.gridSaveSizer.Add(self.SavePlotsBool, pos=(0,0),
                               flag=wx.LEFT, border=5)
        self.gridSaveSizer.Add(self.SaveFitsBool, pos=(1,0),
                               flag=wx.LEFT, border=5)
        self.gridSaveSizer.Add(self.SaveButton, pos=(0,1), span=(2,1),
                               flag=wx.LEFT, border=5)

        for child in self.gridSaveSizer.GetChildren():
            child.GetWindow().Disable()

        # Control buttons
        boxPlotButtonsSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.PlotButton = wx.Button(panel, label='Plot')
        self.PlotAllButton = wx.Button(panel, label='Plot All')
        self.PlotAllButton.Disable()
        boxPlotButtonsSizer.AddMany([(self.PlotButton, flagsCenter),
                                     (self.PlotAllButton, flagsCenter)])


        boxControlButtonsSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.HelpButton = wx.Button(panel, label='Help')
        self.ClearButton = wx.Button(panel, label='Clear')
        self.ClearButton.Disable()
        self.QuitButton = wx.Button(panel, label='Quit')
        boxControlButtonsSizer.AddMany([(self.HelpButton, flagsCenter),
                                     (self.ClearButton, flagsCenter),
                                     (self.QuitButton, flagsCenter)])

        self.PlotButton.Bind(wx.EVT_BUTTON, self.OnPlotPress)
        self.PlotAllButton.Bind(wx.EVT_BUTTON, self.OnPlotAllPress)
        self.HelpButton.Bind(wx.EVT_BUTTON, self.OnHelpPress)
        self.ClearButton.Bind(wx.EVT_BUTTON, self.OnClearPress)
        self.QuitButton.Bind(wx.EVT_BUTTON, self.OnQuitPress)

        #  Add the main sizers to the top sizer
        topsizer.Add(sboxDataSizer, flagsExpand)
        topsizer.Add(sboxConfigSizer, flagsExpand)
        topsizer.Add(self.gridSaveSizer, flagsExpand)
        topsizer.Add(boxPlotButtonsSizer, flagsExpand)
        topsizer.Add(boxControlButtonsSizer, flagsExpand)

        # topsizer.SetSizeHints(self)
        panel.SetSizer(topsizer)
        topsizer.Fit(self)

        # save a zero configuration for convenience
        self.zero_configuration = self.DistConfigSave()
        for key in self.configuration.keys():
            self.configuration[key] = self.zero_configuration.copy()

        # Hold all the parameter widgets in a list for convenience
        self.param_widgets = []

    def DistConfigLoad(self, event):
        # print(self.comboDist.GetValue(), 'Loading configuration')
        dist = self.comboDist.GetValue()

        # load the trace type(s) to be plotted
        for chbox in self.chbTraces:
            chbox.SetValue(self.configuration[dist]['trace'][chbox.GetLabel()])

        for chbox in self.chbSides:
            chbox.SetValue(self.configuration[dist]['side'][chbox.GetLabel()])

        self.entryMinDwell.SetValue(self.configuration[dist]['min'])
        self.entryMaxDwell.SetValue(self.configuration[dist]['max'])

        self.comboScale.SetValue( self.configuration[dist]['scale'])
        self.comboPlotType.SetValue( self.configuration[dist]['PlotType'])
        self.entryBinSize.SetValue(self.configuration[dist]['binsize'])

        self.chbFitBool.SetValue(self.configuration[dist]['FitBool'])
        self.chbTmaxBool.SetValue(self.configuration[dist]['TmaxBool'])
        self.chbBootBool.SetValue(self.configuration[dist]['BootBool'])
        self.comboFitModel.SetValue(self.configuration[dist]['model'])
        self.entryNfits.SetValue(self.configuration[dist]['Nfits'])
        self.entryBoots.SetValue(self.configuration[dist]['BootRepeats'])

        return self.configuration[dist]


    def DistConfigSave(self, event=None):
        dist = self.comboDist.GetValue()
        # print(dist, 'Saving configuration')
        # save the trace type(s) to be plotted
        self.configuration[dist]['trace'] = {}
        for chbox in self.chbTraces:
            self.configuration[dist]['trace'][chbox.GetLabel()] = chbox.GetValue()
        # save the side(s) to be plotted
        self.configuration[dist]['side'] = {}
        for chbox in self.chbSides:
            self.configuration[dist]['side'][chbox.GetLabel()] = chbox.GetValue()

        # save the min and max values
        self.configuration[dist]['min'] = self.entryMinDwell.GetValue()
        self.configuration[dist]['max'] = self.entryMaxDwell.GetValue()

        # save the plot options
        self.configuration[dist]['scale'] = self.comboScale.GetValue()
        self.configuration[dist]['PlotType'] = self.comboPlotType.GetValue()
        self.configuration[dist]['binsize'] = self.entryBinSize.GetValue()

        # save the fit options
        self.configuration[dist]['FitBool'] = self.chbFitBool.GetValue()
        self.configuration[dist]['TmaxBool'] = self.chbTmaxBool.GetValue()
        self.configuration[dist]['BootBool'] = self.chbBootBool.GetValue()

        self.configuration[dist]['model'] = self.comboFitModel.GetValue()
        self.configuration[dist]['Nfits'] = self.entryNfits.GetValue()
        self.configuration[dist]['BootRepeats'] = self.entryBoots.GetValue()

        return self.configuration[dist]

    def onRadioCombinePress(self, event):
        self.entryDataName.Enable()

    def onRadioSeparatePress(self, event):
        self.entryDataName.Disable()

    def OnHelpPress(self, event):
        print('Help')

    def OnPlotPress(self, event):
        print('Plot')
        # self.save_enable()

    def OnPlotAllPress(self, event):
        print('Plot All')
        # self.save_enable()

    def OnClearPress(self, event):
        print('Clear')

    def OnQuitPress(self, event):
        print('Quit')
        self.Hide()

    def save_enable(self):
        for child in self.gridSaveSizer.GetChildren():
            child.GetWindow().Enable()


if __name__ == '__main__':
    app = wx.App()
    ex = Panel(None)
    ex.Show()
    app.MainLoop()




import pytest
import numpy as np
import xarray as xr
import sys

@pytest.fixture
def dataset(shared_datadir):
    return xr.open_dataset(shared_datadir / 'BN_TIRF_output_test_file' / 'TIRF 561 0001.nc' )

def test_trace_plot(dataset):
    from papylio.trace_plot import TracePlotWindow
    frame = TracePlotWindow(dataset)



# -*- coding: utf-8 -*-
"""
Created on Fri Mar  6 01:29:58 2020

@author: iason
"""

import numpy as np

def Exp1(tau, Tmax=1000):
    time = np.linspace(0, Tmax, 1000)
    exp = 1/tau*np.exp(-time/tau)

    return time, exp

def Exp2(p1, tau1, tau2, Tmax=1000):
    time = np.linspace(0, Tmax, 1000)
    exp = p1/tau1*np.exp(-time/tau1)+(1-p1)/tau2*np.exp(-time/tau2)

    return time, exp

def Exp3(p1, p2, tau1, tau2, tau3, Tmax=1000):
    time = np.linspace(0, Tmax, 1000)
    exp = p1/tau1*np.exp(-time/tau1)+p2/tau2*np.exp(-time/tau2) + \
        + (1-p1-p2)/tau3*np.exp(-time/tau3)

    return time, exp


# -*- coding: utf-8 -*-
"""
Created on Mon Jun 22 10:10:04 2020

@author: mwdocter
"""
#import os
#os.getcwd()
#os.chdir('..') # one directory up

import numpy as np
import matplotlib.pyplot as plt
import scipy.ndimage.filters as filters
import math
import logging

logger = logging.getLogger('ftpuploader')
hdlr = logging.FileHandler('ftplog.llog')
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
hdlr.setFormatter(formatter)
logger.addHandler(hdlr)
logger.setLevel(logging.INFO)
FTPADDR = "some ftp address"

# steps in GUI:
from papylio.experiment import Experiment
from papylio.analysis.interactiveAnalysis import InteractivePlot
from papylio.movie.movie import Movie
from papylio.movie.background_correction import get_threshold
from skimage import measure

def autoconfig_AND_perform_mapping(mapping_file_index, main_path):
    exp = Experiment(main_path)
    mapping_file = exp.files[mapping_file_index]
    print(mapping_file.name)
    try : 
        mapping_file.perform_mapping()
        print('no need for autoconfig')
    # if it does not workL autoconfig
    except BaseException as e:
        logger.error('Failed to do something 1: ' + str(e))
        autoconfig(exp, mapping_file, opt='mapping')
        try : 
            #mapping_file.restartreset # doe snot work, full reset of exp works. There should be a simpler way !!!!!
            del exp
            exp = Experiment(main_path)
            mapping_file = exp.files[mapping_file_index]
            mapping_file.perform_mapping()
            print('ran autoconfig once') 
        except BaseException as e:
            logger.error('Failed to do something 2: ' + str(e))
            print('#$%^&*(*&^%$')
     #still not enough points: lower the minimum difference, and raise the factor for mapping
            while True:
                exp.configuration['mapping']['peak_finding']['donor']['minimum_intensity_difference']= \
                  0.9*exp.configuration['mapping']['peak_finding']['donor']['minimum_intensity_difference']
                exp.configuration['mapping']['peak_finding']['acceptor']['minimum_intensity_difference']= \
                  0.9*exp.configuration['mapping']['peak_finding']['acceptor']['minimum_intensity_difference']
                exp.configuration['mapping']['coordinate_optimization']['coordinates_without_intensity_at_radius']['fraction_of_peak_max']=\
                  1.05*exp.configuration['mapping']['coordinate_optimization']['coordinates_without_intensity_at_radius']['fraction_of_peak_max']
                exp.export_config_file()
    
                try : 
                    del exp
                    exp = Experiment(main_path)
                    mapping_file = exp.files[mapping_file_index]
                    mapping_file.perform_mapping()
                    print('ran autoconfig again') 
                    break
                except BaseException as e:
                    logger.error('Failed to do something 3: ' + str(e))
                    if exp.configuration['mapping']['peak_finding']['donor']['minimum_intensity_difference']<1 or\
                       exp.configuration['mapping']['peak_finding']['acceptor']['minimum_intensity_difference']<1 or\
                       exp.configuration['mapping']['coordinate_optimization']['coordinates_without_intensity_at_radius']['fraction_of_peak_max']>1:
                           print('give up, this mapping file is of too low quality')
                           break  
                    continue
               
                
def autoconfig(exp, input_file, opt='mapping'): 
# step 1. find recommended threshold for donor and acceptor channel
    show=0
    average_image = input_file.average_image
    donor_image=input_file.movie.get_channel(channel='d')
    acceptor_image=input_file.movie.get_channel(channel='a')
    thr_donor=get_threshold(donor_image,show=show)
    thr_acceptor=get_threshold(acceptor_image,show=show)
    print('recommended value for mapping - donor - threshold: {0}'.format(thr_donor))
    print('recommended value for mapping - acceptor -  threshold: {0}'.format(thr_acceptor))

# step 2 find recommended thresholds dor the minimum intensity differences for donor&acceptor
    filter_neighbourhood_size=10
    image_max = filters.maximum_filter(donor_image, filter_neighbourhood_size)
    image_min = filters.minimum_filter(donor_image, filter_neighbourhood_size)
    image_diff_donor= image_max-image_min
    image_diff_donor_thr=get_threshold(image_diff_donor)

    image_max = filters.maximum_filter(acceptor_image, filter_neighbourhood_size)
    image_min = filters.minimum_filter(acceptor_image, filter_neighbourhood_size)
    image_diff_acceptor= image_max-image_min
    image_diff_acc_thr=get_threshold(image_diff_acceptor)

    print('recommended value for mapping - donor - min.int.diff: {0}'.format(image_diff_donor_thr))
    print('recommended value for mapping - acceptor - min.int.diff: {0}'.format(image_diff_acc_thr))

# step 3. find the recommended radius and fraction
    label_D2 = measure.label(image_diff_donor>image_diff_donor_thr, 8) 

    imm=donor_image

    D2min=np.zeros([np.max(label_D2), 1])
    D2max=np.zeros([np.max(label_D2), 1])
    area=np.zeros([np.max(label_D2), 1])
    crop_peak=np.zeros((np.max(label_D2), 2))
    crop_peak=np.zeros((np.max(label_D2), 2))
    factors1=np.zeros(np.shape(imm))
    factors0=np.zeros((np.max(label_D2), 2))
    
    # for each labeled spot, find the area, intensity in center and at radius,  
    for ii in range(np.max(label_D2)): 
        if ii>0:
            D2min[ii]=np.min(imm[label_D2==ii])
            D2max[ii]=np.max(imm[label_D2==ii])
            area[ii]=np.sum(label_D2==ii)
            
            location=np.mean(np.where(label_D2==ii),axis=1)
            center = np.round(location[::-1]).astype(int)
            width=4*2+1
            cropped_peak=imm[(center[1]-width//2):(center[1]+width//2+1),(center[0]-width//2):(center[0]+width//2+1)]
            if np.shape(cropped_peak)==(9,9):
                r=4
                d = 2*r + 1
                x, y = np.indices((d, d))
                circle_matrix= (np.abs(np.hypot(r - x, r - y)-r) < 0.5).astype(int)
                cutoff=np.median(imm)
                crop_peak[ii,:]=[np.max(cropped_peak * circle_matrix), np.max(cropped_peak)]
            factors1[label_D2==ii]=(crop_peak[ii,0]-cutoff)/(crop_peak[ii,1]-cutoff)
            factors0[ii,:]=[image_diff_donor[label_D2==ii][0], factors1[label_D2==ii][0]]
    radius=np.ceil(np.sqrt(np.ceil(np.median(area)/(4*np.pi))) )
    print('recommended value for the radius is : {0}'.format(radius) )  

    a,b=np.histogram(factors0[:,1], bins=50)
    factor_threshold= np.sum(b[1:-1]*a[:-1])/np.sum(a[:-1]) #weighted sum, except last entry
    
    print('recommended value for the factor is : {0}'.format(factor_threshold) )  

# step 4, write them in the config file
    print(opt + '\n')
   
    if opt=='mapping':
     #   print('doing mapping')  
        exp.configuration['mapping']['peak_finding']['donor']['minimum_intensity_difference']= \
              int(image_diff_donor_thr  )
        exp.configuration['mapping']['peak_finding']['acceptor']['minimum_intensity_difference']= \
              int(image_diff_acc_thr  )
        exp.configuration['mapping']['coordinate_optimization']['coordinates_without_intensity_at_radius']['radius']=\
              int( radius)
        exp.configuration['mapping']['coordinate_optimization']['coordinates_without_intensity_at_radius']['fraction_of_peak_max']=\
              float(   factor_threshold)
    else:
      #  print('doing find coordinate') 
        exp.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference']= \
              int(image_diff_donor_thr  )
     #   print(exp.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference'])
     #   print(exp.configuration['find_coordinates']['peak_finding']['maximum_intensity_difference'])
          
        if  exp.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference'] > \
            exp.configuration['find_coordinates']['peak_finding']['maximum_intensity_difference']:
        #    print('minimum is above maximum?!')
            exp.configuration['find_coordinates']['peak_finding']['maximum_intensity_difference']=\
            int(image_diff_donor_thr  )+11111
        #    print(exp.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference'])
        #    print(exp.configuration['find_coordinates']['peak_finding']['maximum_intensity_difference'])
       
        exp.configuration['find_coordinates']['coordinate_optimization']['coordinates_without_intensity_at_radius']['radius']=\
               int( radius)
        exp.configuration['find_coordinates']['coordinate_optimization']['coordinates_without_intensity_at_radius']['fraction_of_peak_max']=\
               float( factor_threshold)
    exp.export_config_file()
    


import importlib
import pkgutil

class PluginManager:
    def __init__(self):
        self.plugins_module = importlib.import_module('papylio.plugins')
        self.plugin_names = [pluginname for _, pluginname, ispkg in pkgutil.walk_packages(self.plugins_module.__path__) if ispkg]
        self.plugins = [importlib.import_module('papylio.plugins.'+plugin_name) for plugin_name in self.plugin_names]

    def get_class_plugins(self, class_name):
        return tuple([getattr(plugin, class_name) for plugin in self.plugins if hasattr(plugin, class_name)])


def plugins(cls):
    """
    The function is ment as a decorator for the main classes and makes sure that for the class it is applied to a new class is created with the same name.
    This new class mixes-in the plugin classes with the old class, where the plugin classes have priority and thus more or less inherit from the old class.

    Parameters
    ----------
    cls : type
        The input class

    Returns
    -------
    type
        The adapted class with the same name as the input class, however now with the plugin classes mixed in.

    """
    classes = PluginManager().get_class_plugins(cls.__name__) + (cls,)
    slots = ()
    try:
        for c in classes:
            if type(c.slots) is tuple:
                slots += c.slots
            else:
                slots += (c.slots,)
    except AttributeError:
        pass

    # For using multiprocessing.Pool add '__module__': classes[-1].__module__}
    # return type(cls.__name__, classes, {'__slots__': slots})
    return type(cls.__name__, classes, {'__slots__': slots, '__module__': classes[-1].__module__})



# class PluginMetaClass(type):
#     def __new__(cls, clsname, bases_base, attrs):
#         # bases_base = tuple(base for base in bases if not base.__name__ is clsname)
#         attrs_base = attrs.copy()
#         attrs_base.pop('__qualname__')
#         attrs_base.pop('__module__')
#         cls_base = type(clsname+'_base', bases_base, attrs_base)
#         #cls_base = type(clsname, bases_base, attrs)
#         added_bases = PluginManager().get_class_plugins(clsname)
#         bases_main = added_bases + (cls_base,)
#         return super().__new__(cls, clsname, bases_main, {'__module__': attrs['__module__']})
#
#         # bases_base = tuple(base for base in bases if not base.__name__ is clsname)
#         # attrs.pop('__qualname__')
#         # cls_base = type(clsname + '_base', bases_base, attrs)
#         # bases_main = tuple(base for base in bases if base.__name__ is clsname) + (cls_base,)
#         # return super().__new__(cls, clsname, bases_main, {})
#
#
#
# class B:
#     def __init__(self):
#         print('Badd')
#         super().__init__()
#
#
#
# class A:
#     def __init__(self):
#         print('A')
#
#
# def test(c):
#     return type(c.__name__, (c,B),{})
#
#
# @test
# class Bo(A):
#     def __init__(self):
#         print('Bo')
#         super().__init__()



import numpy as np
import pandas as pd
import xarray as xr
import tifffile

from papylio.movie.movie import Movie
from papylio.timer import Timer


class TifMovie(Movie):
    extensions = ['.tif', '.tiff']

    def __init__(self, arg, *args, **kwargs):
        super().__init__(arg, *args, **kwargs)
        

        self.writepath = self.filepath.parent
        self.name = self.filepath.with_suffix('').name
        
        self.threshold = {  'view':             (0,200),
                            'point-selection':  (45,25)
                            }

        self.file = None # Note this is for the tif file, not the File class.
        self._time = None

        # self.read_header()
        # self.create_frame_info()  # Possibly move to Movie later on

        # self._initialized = True

    def open(self):
        self.file = tifffile.TiffFile(self.filepath)

    def close(self):
        self.file.close()

    def _read_header(self):
        with self:
            tif_tags = {}
            for tag in self.file.pages[0].tags.values():
                name, value = tag.name, tag.value
                tif_tags[name] = value
            self.width = tif_tags['ImageWidth']
            self.height = tif_tags['ImageLength']
            self.data_type = np.dtype(f"uint{tif_tags['BitsPerSample']}")

            if self.file.metaseries_metadata:
                self.number_of_frames = self.file.metaseries_metadata['SetInfo']['number-of-planes']
                # TODO: Make sure this goes well when movie is rotated
                pixel_size_x = self.file.metaseries_metadata['PlaneInfo']['spatial-calibration-x']
                pixel_size_y = self.file.metaseries_metadata['PlaneInfo']['spatial-calibration-y']
                if 'Ti2 Optical Zoom' in self.file.metaseries_metadata['PlaneInfo'].keys():
                    microscope_optical_zoom = float(
                        self.file.metaseries_metadata['PlaneInfo']['Ti2 Optical Zoom'][:-1])
                    pixel_size_x /= microscope_optical_zoom
                    pixel_size_y /= microscope_optical_zoom
                self.pixel_size = np.array([pixel_size_x, pixel_size_y])
                self.pixel_size_unit = self.file.metaseries_metadata['PlaneInfo'][
                    'spatial-calibration-units'].replace('um', 'µm')
                # TODO: Make sure this goes well when movie is rotated
                stage_position_x = self.file.metaseries_metadata['PlaneInfo']['stage-position-x']
                stage_position_y = self.file.metaseries_metadata['PlaneInfo']['stage-position-y']
                self.stage_coordinates = np.array([[stage_position_x, stage_position_y]])
                self.stage_coordinates_in_pixels = self.stage_coordinates / self.pixel_size
            else:
                self.number_of_frames = len(self.file.pages)
                self.pixel_size = np.array([1, 1])  # For TIR-T because it is not in metadata
                self.pixel_size_unit = ''  # For TIR-T because it is not in metadata
                # Or self.file.series[0].shape[0]

            # if 'DateTime' in self.file.pages[0].tags:
            #     # For which files is this?
            #     # Note that looping over tif pages will be slow.
            #     self.datetime = pd.to_datetime([page.tags['DateTime'].value for page in self.file.pages])
            #     self.time = xr.DataArray((self.datetime-self.datetime[0]).total_seconds(), dims='frame',
            #                              coords={}, attrs={'units': 's'})
            # elif 'DateTime' in self.file.pages[0].tags:
            #     # This is for tiff images from TIR-T and TIR-V measured by Solis software
            #     exposure_time = self.file.pages[0].tags['AndorExposureTime'].value # NOTE: "kinetic time (or cycle time)" is more accurate measure. But the difference is very small and often ignored.
            #     time_vector = exposure_time * np.arange(0, self.number_of_frames)
            #     self.time = xr.DataArray(time_vector, dims='frame', coords={}, attrs={'units': 's'})
    #
    @property
    def time(self):
        # TODO: Get time when in read_frame
        if self._time is None:
            with self:
                # print('Time is obtained')
                if 'DateTime' in self.file.pages[0].tags:
                    # For which files is this?
                    # Note that looping over tif pages will be slow.
                    self.datetime = pd.to_datetime([page.tags['DateTime'].value for page in self.file.pages])
                    self._time = xr.DataArray((self.datetime-self.datetime[0]).total_seconds(), dims='frame',
                                             coords={}, attrs={'units': 's'})
                elif 'AndorExposureTime' in self.file.pages[0].tags:
                    # This is for tiff images from TIR-T and TIR-V measured by Solis software
                    exposure_time = self.file.pages[0].tags['AndorExposureTime'].value # NOTE: "kinetic time (or cycle time)" is more accurate measure. But the difference is very small and often ignored.
                    time_vector = exposure_time * np.arange(0, self.number_of_frames)
                    self._time = xr.DataArray(time_vector, dims='frame', coords={}, attrs={'units': 's'})

                # self.create_frame_info()  # Possibly move to Movie later on
        return self._time

    def _read_frame(self, frame_number):
        with self:
            # if self.number_of_frames == 1:
            #     # return -1,0,0,0
            #     im = tifpage[0].asarray()
            if (self.number_of_frames - 1) >= frame_number:
                im = self.file.pages[frame_number].asarray().astype(self.data_type)
            else:
                raise IndexError('Selected frame number larger than number of frames')
            #     im = tifpage[self.number_of_frames - 1].asarray()
            #     print('pageNb out of range, printed image {0} instead'.format(self.number_of_frames))
        return im

    def _read_frames(self, frame_indices=None):
        if not self.use_dask:
            if len(frame_indices) == 1:
                with self:
                    frames = np.stack([self.file.pages[i].asarray() for i in frame_indices])
            else:
                frames = tifffile.imread(self.filepath, key=frame_indices)

        else:
            import dask_image.imread
            frames = dask_image.imread.imread(self.filepath, nframes=100)[frame_indices]
        return frames

if __name__ == "__main__":
    movie = TifMovie(r'.\Example_data\tif\movie.tif')

#
# import os
# from glob import glob
#
# try:
#     from skimage.io import imread as sk_imread
# except (AttributeError, ImportError):
#     pass
#
# from dask.base import tokenize
# from dask.array.core import Array
#
#
# def add_leading_dimension(x):
#     return x[None, ...]
#
#
# def imread(filename, imread=None, preprocess=None):
#     """Read a stack of images into a dask array
#
#     Parameters
#     ----------
#
#     filename: string
#         A globstring like 'myfile.*.png'
#     imread: function (optional)
#         Optionally provide custom imread function.
#         Function should expect a filename and produce a numpy array.
#         Defaults to ``skimage.io.imread``.
#     preprocess: function (optional)
#         Optionally provide custom function to preprocess the image.
#         Function should expect a numpy array for a single image.
#
#     Examples
#     --------
#
#     >>> from dask.array.image import imread
#     >>> im = imread('2015-*-*.png')  # doctest: +SKIP
#     >>> im.shape  # doctest: +SKIP
#     (365, 1000, 1000, 3)
#
#     Returns
#     -------
#
#     Dask array of all images stacked along the first dimension.  All images
#     will be treated as individual chunks
#     """
#     file = tifffile.TiffFile(filename)
#
#     imread = imread or sk_imread
#     filenames = sorted(glob(filename))
#     if not filenames:
#         raise ValueError("No files found under name %s" % filename)
#
#     name = "imread-%s" % tokenize(file.pages, map(os.path.getmtime, file.pages))
#
#     sample = file.pages[0].asarray()
#     if preprocess:
#         sample = preprocess(sample)
#
#     keys = [(name, i) + (0,) * len(sample.shape) for i in range(len(file.pages))]
#     if preprocess:
#         values = [
#             (add_leading_dimension, (tifffile.TiffPage.asarray, page)) for page in file.pages
#         ]
#     else:
#         values = [(add_leading_dimension, (tifffile.TiffPage.asarray, page)) for page in file.pages]
#     dsk = dict(zip(keys, values))
#
#     chunks = ((1,) * len(file.pages),) + tuple((d,) for d in sample.shape)
#
#     return Array(dsk, name, chunks, sample.dtype)


import pytest
import tifffile
import numpy as np


@pytest.fixture
def experiment(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF_sequencing')

@pytest.fixture
def file(experiment):
    return experiment.files[1]


def test_sequencing_dataset_property_in_file(file):
    sequencing_data = file.sequencing_data
    sequencing_data.dataset.attrs['test'] = 'test'
    file.sequencing_data = sequencing_data
    assert file.sequencing_data.attrs['test'] == 'test'


def test_get_sequencing_data(file):
    file.get_sequencing_data(margin=5, mapping_name='All files')


def test_generate_sequencing_match(file):
    file.generate_sequencing_match(overlapping_points_threshold=25, excluded_sequence_names=['MapSeq', 'CalSeq', '*'])


def test_insert_sequencing_data_into_file_dataset(file):
    file.insert_sequencing_data_into_file_dataset()


# -*- coding: utf-8 -*-
"""
Created on Fri Dec  6 18:59:11 2019

@author: pimam
"""
if __name__ == '__main__':
    import os
    import sys
    from pathlib import Path, PureWindowsPath
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    p = Path(__file__).parents[3]
    sys.path.insert(0, str(p))
    main_path = PureWindowsPath('C:\\Users\\pimam\\Documents\\MEP\\tracesfiles')

import numpy as np
import pandas as pd
import matplotlib.pylab as plt
import seaborn as sns
sns.set(style="dark")
sns.set_color_codes()


def ML1expcut(dwells, Tcut, Ncut):
    if Ncut == 0:
        MLtau = np.average(dwells)
    else:
        Nrec = dwells.size
        avg_dwells = np.average(dwells)
        MLtau = avg_dwells + Ncut*Tcut/Nrec
    timearray = np.linspace(0, Tcut, 1000)
    P = 1/MLtau*np.exp(-timearray/MLtau)
    return P, MLtau


def P2expcut(dwells, params, Tcut, Ncut):
    P1, tau1, tau2 = params
    Pi = P1/tau1*np.exp(-dwells/tau1)+(1-P1)/tau2*np.exp(-dwells/tau2)
    Pcut = P1*np.exp(-Tcut/tau1)+(1-P1)*np.exp(-Tcut/tau2)
    return Pi, Pcut

def P3expcut(dwells, params, Tcut, Ncut):
    P1, P2, tau1, tau2, tau3 = params
    Pi = P1/tau1*np.exp(-dwells/tau1)+P2/tau2*np.exp(-dwells/tau2) + \
        (1 - P1 - P2)/tau3*np.exp(-dwells/tau3)
    Pcut = P1*np.exp(-Tcut/tau1)+P2*np.exp(-Tcut/tau2) + \
        (1 - P1 - P2)*np.exp(-Tcut/tau3)
    return Pi, Pcut

def BIC(dwells, k, LogLike):
    bic = np.log(dwells.size)*k - 2*LogLike
    return bic


def LogLikelihood(dwells, params, model, Tcut, Ncut):
    Pi, Pcut = model(dwells, params, Tcut, Ncut)

    LLikecut = -Ncut * np.log(Pcut)
    LLike = np.sum(-np.log(Pi)) + LLikecut
    return LLike


def update_temp(T, alpha):
    '''
    Exponential cooling scheme
    :param T: current temperature.
    :param alpha: Cooling rate.
    :return: new temperature
    '''
    T *= alpha
    return T


def simulated_annealing(data, objective_function, model, x_initial, lwrbnd,
                         uprbnd, Tcut, Ncut, Tstart=100.,
                         Tfinal=0.001, delta1=0.1, delta2=2.5, alpha=0.9):
    i = 0
    T = Tstart
    step = 0
    xstep = 0
    x = x_initial
    while T > Tfinal:
        step += 1
        if (step % 100 == 0):
            T = update_temp(T, alpha)
        x_trial = np.zeros(len(x))
        x_trial[0] = np.random.uniform(np.max([x[0] - delta1, lwrbnd[0]]),
                                       np.min([x[0] + delta1, uprbnd[0]]))
        for i in range(1, len(x)):
            x_trial[i] = np.random.uniform(np.max([x[i] - delta2, lwrbnd[i]]),
                                           np.min([x[i] + delta2, uprbnd[i]]))
        x, xstep = Metropolis(objective_function, model, x, x_trial, T, data,
                              Tcut, Ncut, xstep)

    return x, xstep


def Metropolis(f, model, x, x_trial, T, data, Tcut, Ncut, xstep):
    # Metropolis Algorithm to decide if you accept the trial solution.
    Vnew = f(data, x_trial, model, Tcut, Ncut)
    Vold = f(data, x, model, Tcut, Ncut)
    if (np.random.uniform() < np.exp(-(Vnew - Vold) / T)):
        x = x_trial
        xstep += 1
    return x, xstep


def Best_of_Nfits_sim_anneal(dwells, Nfits, model, x_initial,
                             lwrbnd, uprbnd, Tcut, Ncut):
    # Perform N fits on data using simmulated annealing
    LLike = np.empty(Nfits)
    for i in range(0, Nfits):
        fitdata, xstep = simulated_annealing(data=dwells,
                                              objective_function=LogLikelihood,
                                              model=model, x_initial=x_initial,
                                              lwrbnd=lwrbnd, uprbnd=uprbnd,
                                              Tcut=Tcut, Ncut=Ncut)
        # print(f"fit{i} found: {fitdata}")
        if i == 0:
            fitparam = [fitdata]
            Nsteps = [xstep]
        else:
            fitparam = np.concatenate((fitparam, [fitdata]), axis=0)
            Nsteps = np.concatenate((Nsteps, [xstep]), axis=0)
        LLike[i] = LogLikelihood(dwells, fitparam[i], model, Tcut, Ncut)
    ibestparam = np.argmax(LLike)
    bestparam = fitparam[ibestparam]
    bestNsteps = Nsteps[ibestparam]
    return bestparam, bestNsteps

def Bootstrap_data(dwells, Ncut):
    dwells_Ncut = np.concatenate((dwells, np.zeros(Ncut)))
    dwells_rand = np.random.choice(dwells_Ncut, dwells_Ncut.size)
    Bootstrapped_dwells = dwells_rand[dwells_rand > 0]
    Bootstrapped_Ncut = dwells_rand[dwells_rand == 0].size
    return Bootstrapped_dwells, Bootstrapped_Ncut


def fit(dwells_all, mdl, dataset_name='Dwells', Nfits=1,
        include_over_Tmax=True, bootstrap=False, boot_repeats=0):
    Tmax = dwells_all.max()
    if include_over_Tmax:
        Tmax = Tmax - 5
        dwells = dwells_all[dwells_all < Tmax]
        Ncut = dwells_all[dwells_all >= Tmax].size
        print(f'Ncut: {Ncut}')
    else:
        Ncut = 0
        dwells = dwells_all

    # the initial holder for the fit result irrespective of the fit model
    fit_result = pd.DataFrame({'Dataset': [dataset_name], 'model': [mdl]})


    if mdl == '1Exp':
        #  The 1exp fit is given by analytic solution, just the average dwelltime
        fit, bestvalue = ML1expcut(dwells, Tmax, Ncut)
        error = 0
        boot_params = np.empty(boot_repeats)
        if bootstrap is True:
            Ncutarray = np.empty(boot_repeats)
            for i in range(0, boot_repeats):
                boot_dwells, boot_Ncut = Bootstrap_data(dwells, Ncut)
                fit_boot, param = ML1expcut(boot_dwells, Tmax, boot_Ncut)
                Ncutarray[i] = boot_Ncut
                boot_params[i] = param

            error = np.std(boot_params)

        result = pd.DataFrame({'param': ['tau'], 'value': [bestvalue],
                               'error': [error], 'Tmax': [Tmax], 'Ncut': [Ncut],
                               'BootRepeats': [boot_repeats*bootstrap],
                               'steps': ['N/A']})

        fit_result = pd.concat([fit_result, result], axis=1)


    elif mdl == '2Exp':
        # For 2exp fit the maximum likelihood of the 2exp model is obtained with
        # simulated annealing minimization of -log(ML)
        model = P2expcut

        # Set parameters for simmulated annealing
        avg_dwells = np.average(dwells)
        x_initial = [0.5, avg_dwells, avg_dwells]
        lwrbnd = [0, 0, 0]
        uprbnd = [1, 3*Tmax, 3*Tmax]

        # Perform N fits on data using simmulated annealing and select best
        bestvalues, bestNsteps = Best_of_Nfits_sim_anneal(
                                                        dwells, Nfits,
                                                        model=model,
                                                        x_initial=x_initial,
                                                        lwrbnd=lwrbnd,
                                                        uprbnd=uprbnd,
                                                        Tcut=Tmax,
                                                        Ncut=Ncut)


        # make sure the fit parameters are ordered from low to high dwelltimes
        if bestvalues[1] > bestvalues[2]:
            bestvalues = [1-bestvalues[0]] + [bestvalues[2], bestvalues[1]]

        errors = [0, 0, 0]
        boot_params = np.empty((boot_repeats,3))
        # Check if bootstrapping is used
        if bootstrap:
            LLike = np.empty(boot_repeats)
            Ncutarray = np.empty(boot_repeats)
            Nstepsarray = np.empty(boot_repeats)
            print('bootrepeats: ', boot_repeats)
            for i in range(0, boot_repeats):
                boot_dwells, boot_Ncut = Bootstrap_data(dwells, Ncut)
                params, Nsteps =simulated_annealing(
                                                    boot_dwells,
                                                    LogLikelihood,
                                                    model=model,
                                                    x_initial=x_initial,
                                                    lwrbnd=lwrbnd,
                                                    uprbnd=uprbnd,
                                                    Tcut=Tmax,
                                                    Ncut=boot_Ncut)
                print(f'boot: {i+1}, steps: {Nsteps}')
                # make sure the fit parameters are ordered from low to high dwelltimes
                if params[1] > params[2]:
                    params = [1-params[0]] + [params[2], params[1]]

                Ncutarray[i] = boot_Ncut
                Nstepsarray[i] = Nsteps
                boot_params[i] = params
                LLike[i] = LogLikelihood(dwells,params, model, Tmax, Ncut)
            errors = np.std(boot_params, axis=0)

        # Put fit result into dataframe

        result = pd.DataFrame({'param': ['p', 'tau1', 'tau2'],
                              'value': bestvalues, 'error': errors})

        # Calculate the BIC
        LogLike = LogLikelihood(dwells, bestvalues, model, Tmax, Ncut)
        bic = BIC(dwells, len(bestvalues), LogLike)

        result_rest = pd.DataFrame({'Tmax': [Tmax], 'Ncut': [Ncut],
                               'BootRepeats': [boot_repeats*bootstrap],
                               'steps': [bestNsteps], 'BIC': bic})

        fit_result = pd.concat([fit_result, result, result_rest], axis=1)


    elif mdl == '3Exp':
        # For 2exp fit the maximum likelihood of the 2exp model is obtained with
        # simulated annealing minimization of -log(ML)
        model = P3expcut

        # Set parameters for simmulated annealing
        avg_dwells = np.average(dwells)
        x_initial = [0.5, 0.5, avg_dwells, avg_dwells, avg_dwells]
        lwrbnd = [0, 0, 0, 0, 0]
        uprbnd = [1, 1, 3*Tmax, 3*Tmax, 3*Tmax]

        # Perform N fits on data using simmulated annealing and select best
        bestvalues, bestNsteps = Best_of_Nfits_sim_anneal(
                                                        dwells, Nfits,
                                                        model=model,
                                                        x_initial=x_initial,
                                                        lwrbnd=lwrbnd,
                                                        uprbnd=uprbnd,
                                                        Tcut=Tmax,
                                                        Ncut=Ncut)

        # make sure the fit parameters are ordered from low to high dwelltimes
        # if bestvalues[1] > bestvalues[2]:
        #     bestvalues = [1-bestvalues[0]] + [bestvalues[2], bestvalues[1]]

        errors = [0, 0, 0, 0, 0]
        boot_params = np.empty((boot_repeats,5))
        # Check if bootstrapping is used
        if bootstrap:
            LLike = np.empty(boot_repeats)
            Ncutarray = np.empty(boot_repeats)
            Nstepsarray = np.empty(boot_repeats)
            print('bootrepeats: ', boot_repeats)
            for i in range(0, boot_repeats):
                boot_dwells, boot_Ncut = Bootstrap_data(dwells, Ncut)
                params, Nsteps =simulated_annealing(
                                                    boot_dwells,
                                                    LogLikelihood,
                                                    model=model,
                                                    x_initial=x_initial,
                                                    lwrbnd=lwrbnd,
                                                    uprbnd=uprbnd,
                                                    Tcut=Tmax,
                                                    Ncut=boot_Ncut)
                print(f'boot: {i+1}, steps: {Nsteps}')
                # make sure the fit parameters are ordered from low to high dwelltimes
                # if params[1] > params[2]:
                #     params = [1-params[0]] + [params[2], params[1]]

                Ncutarray[i] = boot_Ncut
                Nstepsarray[i] = Nsteps
                boot_params[i] = params
                LLike[i] = LogLikelihood(dwells,params, model, Tmax, Ncut)
            errors = np.std(boot_params, axis=0)

        # Put fit result into dataframe

        result = pd.DataFrame({'param': ['p1', 'p2', 'tau1', 'tau2', 'tau3'],
                              'value': bestvalues, 'error': errors})
        # Calculate the BIC
        LogLike = LogLikelihood(dwells, bestvalues, model, Tmax, Ncut)
        bic = BIC(dwells, len(bestvalues), LogLike)

        result_rest = pd.DataFrame({'Tmax': [Tmax], 'Ncut': [Ncut],
                               'BootRepeats': [boot_repeats*bootstrap],
                               'steps': [bestNsteps], 'BIC': bic})

        fit_result = pd.concat([fit_result, result, result_rest], axis=1)

    return fit_result, boot_params


if __name__ == '__main__':

    # Import data and prepare for fitting
    path = 'C:/Users/iason/Desktop/traceanalysis/trace_analysis/traces/'
    filename = 'hel0_dwells_data.xlsx'
    dwells_all = pd.read_excel(path+filename).offtime.dropna().values

    # Start fitting
    mdl = '2Exp'

    include_over_Tmax = False
    Nfits = 1
    bootstrap = True
    boot_repeats = 10
    result, boot = fit(dwells_all, mdl, 'test', Nfits, include_over_Tmax,
                  bootstrap, boot_repeats)
    print(result)
    plt.hist(boot)
    # if bootstrap is True:
    #     fitdata.to_csv(f'{mdl}_inclTmax_{include_over_Tmax}_bootstrap{boot_repeats}.csv', index=False)
    # else:
    #     fitdata.to_csv(f'{mdl}_inclTmax_{include_over_Tmax}_Nfits{Nfits}.csv', index=False)

#    newdata = pd.read_csv(f'{mdl}_inclTmax_{include_over_Tmax}_bootstrap{boot_repeats}.csv')

    # Getting measures and plotting the parameter values found
    # taubnd = 100
    # fitP1 = []
    # fittau1 = []
    # fittau2 = []
    # for i in range(0, len(fitdata['tau1'])):
    #     if fitdata['tau1'][i] > taubnd:
    #         fittau2.append(fitdata['tau1'][i])
    #         fitP1.append(1-fitdata['P1'][i])
    #     else:
    #         fittau1.append(fitdata['tau1'][i])
    #         fitP1.append(fitdata['P1'][i])
    #     if fitdata['tau2'][i] > taubnd:
    #         fittau2.append(fitdata['tau2'][i])
    #     else:
    #         fittau1.append(fitdata['tau2'][i])

    # P1_avg = np.average(fitP1)
    # tau1_avg = np.average(fittau1)
    # tau2_avg = np.average(fittau2)
    # P1_std = np.std(fitP1)
    # tau1_std = np.std(fittau1)
    # tau2_std = np.std(fittau2)
    # Nbins = 50

    # plt.figure()
    # plt.hist(fitP1, bins=Nbins)
    # plt.vlines(P1_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(P1_avg))
    # plt.title(f'Fit values for P1 Nfits: {boot_repeats} Nbins: {Nbins}')
    # plt.legend()
    # plt.figure()
    # plt.hist(fittau1, bins=Nbins)
    # plt.vlines(tau1_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(tau1_avg))
    # plt.title(rf'Fit values for $\tau$1 Nfits: {boot_repeats} Nbins: {Nbins}')
    # plt.legend()
    # plt.figure()
    # plt.hist(fittau2, bins=Nbins)
    # plt.vlines(tau2_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(tau2_avg))
    # plt.title(rf'Fit values for $\tau$1 Nfits: {boot_repeats} Nbins: {Nbins}')
    # plt.legend()


#    # Plot data with double and single exponential fit
#    plt.figure()
#    plt.semilogy(centers, values, '.', label=f'Dwells with Ncut:{Ncut}')
#    plt.semilogy(timearray, bestfit, label='P1:'+"{0:.2f}".format(bestparams[0])+"\n"+r'$\tau$1:'+"{0:.1f}".format(bestparams[1])+"\n"+r'$\tau$2:'+"{0:.1f}".format(bestparams[2]))
#    singlexp = 1/avg_dwells*np.exp(-timearray/avg_dwells)
#    plt.plot(timearray, singlexp, 'orange', label = rf'$\tau$:{avg_dwells:.1f}')
#    plt.xlabel('dwell time (sec)')
#    plt.ylabel('log prob. density')
#    plt.legend(fontsize='x-large')
#  #  plt.savefig(f'{len(exp.files)}files_1_2expfit__compared.png', dpi=200)


# def ML2expcut(dwells, params, Tcut, Ncut):  # not used
#     P1, tau1, tau2 = params
#     Pi = P1/tau1*np.exp(-dwells/tau1)+(1-P1)/tau2*np.exp(-dwells/tau2)
#     LLike = np.sum(-np.log(Pi))
#     if Ncut != 0:
#         Pcut = P1*np.exp(-Tcut/tau1)+(1-P1)*np.exp(-Tcut/tau2)
#         LLikecut = -Ncut * np.log(Pcut)
#         LLike += LLikecut
#     return Pi, LLike


from papylio.movie.movie import Movie


def test_image_info_from_filename():
    # movie = Movie('Abc.tif')
    # movie.number_of_frames = 100
    filename = 'Abc_fov005_ave_i0_f10-50-2.tif'
    image_info = Movie.image_info_from_filename(filename)
    image_info_expected = {'fov_index': 5, 'image_type': 'average', 'frame_range': (10, 50, 2),
                           'illumination_index': 0}
    assert image_info == image_info_expected


def test_image_info_to_filename():
    # movie = Movie('Abc.tif')
    # movie.number_of_frames = 100
    # movie.illumination_arrangement = [0, 1]
    image_info = {'fov_index': 5, 'image_type': 'average', 'frame_range': (10, 50, 2),
                  'illumination_index': 0}
    filename = Movie.image_info_to_filename(**image_info)
    filename_expected = 'Abc_ave_fov005_f10-50-2_i0'
    assert filename == filename_expected
